
"""
Name: 'The Movies Animation v277.1 (.anm)'
Blender: 277
Group: 'Import'
Tooltip: 'Import from Lionhead Studios The Movies Animation file format. (.anm)'
"""

__author__ = ["Glen Rickey"]
__url__ = ("Director's Cut Modding Foundry","www.dcmodding.com")
__version__ = "1.00 05-12-2009"
__bpydoc__ = """\

Anm Importer

This script imports an anm file from Lionhead Studio Limited's The Movies
into blender for editing. It is not supported by either Lionhead or
Activision.

Known Problems:

Expects to be imported into an existing TM mesh that has been imported with the proper armature bone structure and/or correct group_names.

Modified to run in Blender 2.77, 9/19/2018 by Cary Semar (lefty2000).

Modified:
    5/3/2022
    Changed code to compute "bonelist" data from armature.
    5/6/2022
    Changed code to be compatible with blender 2.49b mesh imports.
    This version checks armature properties to see if 'BoneOrder'
    property is present.  If not, the bone order data is loaded
    from arm['TheMovies']['BoneOrder']
    

"""

# ***** BEGIN GPL LICENSE BLOCK *****
#
# Script copyright (C) Mark S Andrews, Nick Hudson, Glen Rickey, Cary Semar
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****
# --------------------------------------------------------------------------


# Import Modules
#import Blender
#from Blender import Image, Material, Texture, Window, Armature
#from Blender import Mesh, NMesh, Group
#from Blender import Ipo
#from Blender.Mathutils import Matrix, Vector, LineIntersect, DotVecs, AngleBetweenVecs, TranslationMatrix
import struct, os, math, time
import bpy,mathutils
from bpy import context
from mathutils import Matrix,Vector,Quaternion,Euler
#import tmConst
#import pdb
print('Installed: tm_anm_import 3.4')
AH_CURRENT_VERSION = 12
compression = []
origins = []
def UncompressU16(val):
    return (val/32767.0 - 1)
#return (val/32767.0 )
#
# file selector
class TheMoviesAnmImporter(bpy.types.Operator):
    """Import a TM animation"""
    bl_idname = "object.animation_import"
    bl_label = "Import"    
 
    filepath = bpy.props.StringProperty(subtype="FILE_PATH")
 
    my_float = bpy.props.FloatProperty(name="Float")
    my_bool = bpy.props.BoolProperty(name="Toggle Option")
    my_string = bpy.props.StringProperty(name="String Value")
 
    def execute(self, context):
        load_anm(self.filepath)
        return {'FINISHED'}
 
    def invoke(self, context, event):
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}
 
    def draw(self, context):
        layout = self.layout
        col = layout.column()
        col.label(text="Custom Interface!")
 
        row = col.row()
        row.prop(self, "my_float")
        row.prop(self, "my_bool")
 
        col.prop(self, "my_string")
 
def menu_func_import(self,context):
    self.layout.operator(TheMoviesAnmImporter.bl_idname, text = "Import The Movies Animation (*.anm)")
def register():
    bpy.utils.register_class(TheMoviesAnmImporter)
    bpy.types.INFO_MT_file_import.append(menu_func_import)

#***********************************************
# helper classes & functions for files
#***********************************************
#  Orieintaton of root bone's parent
mr = Matrix([[0.,0.,1.],
             [1.,0.,0.],
             [0.,1.,0.]])
mrt = Matrix([[0.,1.,0.],
              [0.,0.,1.],
              [1.,0.,0.]])
bns = []
bqr = []
locs = []
bplist = {}
bone_prent_list = []
arms = []
animorigins = []
def read_item(struct_type, stream):
    value = struct.unpack(struct_type,
    stream.read(struct.calcsize(struct_type)))
    if len(value) == 1:
        value = value[0]
    return value

def read_nts(stream, length=32):
    s = stream.read(length)
#   print('read_nts = ',stream.tell())
#    return s
    return s.decode("ASCII")

def pad(stream):
    align = stream.tell()
    if ((align % 4)>0):
        #skip padding to next 4-byte boundary
        stream.read(4 - (align % 4))

class decoder(object):
    _struct = ''
    _fields = {}
    _flags = {}

    def __init__(self, stream):
        self._value = struct.unpack(self._struct,
        stream.read(struct.calcsize(self._struct)))

    def __getattr__(self, item):
        if item in self._fields.keys():
            keyid = self._fields[item]
            return self._value[keyid]
        elif item in self._flags.keys():
            flag, bit = self._flags[item]
            return bool(self._value[flag] & bit)
        else:
            raise AttributeError(item)

    def __repr__(self):
        return repr(self._value)

#***********************************************
# class definitions for anm file
#***********************************************

class anm_header(decoder):
    _struct = "LBBHffffffffffffBxxx"
    _fields = {
        "Magic":0,
        "count_anims":1,
        "count_static_anims":2,
        "count_frames":3,
        "m11":4,
        "m12":5,
        "m13":6,
        "m21":7,
        "m22":8,
        "m23":9,
        "m31":10,
        "m32":11,
        "m33":12,
        "offset_x":13,
        "offset_y":14,
        "offset_z":15,
        }
    _flags = {
        "is_variance":(16,1),
        "is_30fps": (16,2)
        }
    def __init__(self, stream):
        decoder.__init__(self, stream)
        m1 = [self.m11, self.m12, self.m13,0]
        m2 = [self.m21, self.m22, self.m23,0]
        m3 = [self.m31, self.m32, self.m33,0]
        m4 = [self.offset_x, self.offset_y, self.offset_z,0]
        self.matrix = Matrix()
        self.matrix[0] = Vector(m1)
        self.matrix[1] = Vector(m2)
        self.matrix[2] = Vector(m3)
        self.matrix[3] = Vector(m4)
#        self.matrix = Matrix(m1, m2, m3, m4).resize4x4()

class anm_origin(decoder):
    _struct = "ffffffffffff"
    _fields = {
        "m11":0,
        "m12":1,
        "m13":2,
        "m21":3,
        "m22":4,
        "m23":5,
        "m31":6,
        "m32":7,
        "m33":8,
        "offset_x":9,
        "offset_y":10,
        "offset_z":11,
    }
    def __init__(self, stream):
        global origins
        decoder.__init__(self, stream)
        m1 = [self.m11, self.m12, self.m13,0]
        m2 = [self.m21, self.m22, self.m23,0]
        m3 = [self.m31, self.m32, self.m33,0]
        m4 = [self.offset_x, self.offset_y, self.offset_z,0]
        self.matrix = Matrix()
        self.matrix[0] = Vector(m1)
        self.matrix[1] = Vector(m2)
        self.matrix[2] = Vector(m3)
        self.matrix[3] = Vector(m4)
        self.origin = self.matrix
        origins += [self.origin]
class bone_anm_header(decoder):
    _struct = "LHBBffff"
    _fields = {
        "frame_count": 0,
        "bone_count": 1,
        "besson_type": 2,
        "dest_x": 4,
        "dest_y": 5,
        "dest_z": 6,
        "dest_angle": 7
        }
    _flags = {
        "has_scale": (3,1),
        "has_probability": (3,2),
        "has_animated_scale": (3,4)
        }
    def __init__(self, stream):
        global compression
        decoder.__init__(self, stream)
        if self.has_probability:
            self.probability = read_item("f", stream)
        if self.has_scale:
            self.compression = read_item("f", stream)
            compression += [self.compression]

class BoneFrames(list):
    def __init__(self, stream, framecount):
        #bf = [None] * framecount
        bf = []
        mask = read_item("H", stream)
        fc = framecount
        while fc > 0:
            holdcount = 1
            data = read_item("H", stream)
            if (data & mask):
                v1 = data & (mask - 1)
                v2 = data - mask
                v3 = ~(mask - 1)
                holdcount = v1 | ((v2 & v3) >> 1)
                if (holdcount>framecount) or (holdcount<0):
#                    raise ValueError, 'value is outside bounds for hold.'
                    raise ValueError('value is outside bounds for hold')
                holdvalue = UncompressU16(read_item("H", stream))
                for hold in range(holdcount):
                    bf.append(holdvalue)
            else:
                bf.append(UncompressU16(data))
            fc -= holdcount
        super(BoneFrames, self).__init__(bf)

class BoneAnim(object):
    def __init__(self, stream, framecount):
        self.trans_static = read_item("bxxx", stream)
        self.x = BoneFrames(stream, framecount)
        self.y = BoneFrames(stream, framecount)
        self.z = BoneFrames(stream, framecount)
        pad(stream)
        if self.trans_static == 0:
            self.u = BoneFrames(stream, framecount)
            self.v = BoneFrames(stream, framecount)
            self.w = BoneFrames(stream, framecount)
        else:
            self.u = [read_item("f", stream)] * framecount
            self.v = [read_item("f", stream)] * framecount
            self.w = [read_item("f", stream)] * framecount
        pad(stream)

class BoneAnims(list):
    def __init__(self, stream):
        self.header = bone_anm_header(stream)
        self.bone_list = []
        for b in range(self.header.bone_count):
            self.bone_list.append(read_item("b", stream))
        pad(stream)
        ba_list = []
        for ba in range(self.header.bone_count):
            ba_list.append(BoneAnim(stream, self.header.frame_count))

        BoneAnms = zip(self.bone_list, ba_list)
        super(BoneAnims, self).__init__(BoneAnms)


class prop_anm_header(object):
    def __init__(self, stream):
        val = read_item("Hxx", stream)
        self.is_carriable = val & 1
        self.has_scale = (val & 2)>>1
        self.index_like_me = ((val | 32707) - 32707) >> 2 #This is 4 bits wide.. 0-15 valid values
        self.is_more_than_one_like_me = (val & 64) >> 6
        self.index_skeleton = (val | 127) >> 7 #This is the remainder of 8 bits wide...0-255 valid values
        if self.has_scale:
            self.compression = read_item("f", stream)

class PropFrame(object):
    def __init__(self, stream):
        self.x = UncompressU16(read_item("H",stream))
        self.y = UncompressU16(read_item("H",stream))
        self.z = UncompressU16(read_item("H",stream))
        self.u = UncompressU16(read_item("H",stream))
        self.v = UncompressU16(read_item("H",stream))
        self.w = UncompressU16(read_item("H",stream))

class PropFrames(list):
    def __init__(self, stream, framecount):
        PropPos = [None] * framecount
        for j in range(framecount):
            PropPos[j] = PropFrame(stream)
        super(PropFrames, self).__init__(PropPos)


class PropAnim(object):
    def __init__(self, stream, framecount):
        self.name = read_nts(stream)
#        print('PropAnim: self.name = ',self.name)
        self.header = prop_anm_header(stream)
        self.frames = PropFrames(stream, framecount)




#***********************************************
def quat_mult(r,q):
    p0 = r[0]*q[0]-r[1]*q[1]-r[2]*q[2]-r[3]*q[3]
    p1 = r[0]*q[1]+r[1]*q[0]-r[2]*q[3]+r[3]*q[2]
    p2 = r[0]*q[2]+r[1]*q[3]+r[2]*q[0]-r[3]*q[1]
    p3 = r[0]*q[3]-r[1]*q[2]+r[2]*q[1]+r[3]*q[0]
    p = mathutils.Quaternion((p0,p1,p2,p3))
    return p
def generate_qbrs(arm):
    global bns,bqr,locs,bplist,bone_parent_list
    bns = []
    bqr = []
    locs = []
    scn = bpy.data.scenes[0]
#    arm = scn.objects.active
#for b in arm.data.bones:
    if 'BoneOrder' in arm:
        D = arm['BoneOrder']
    else:
        D = arm['TheMovies']['BoneOrder']
    for key in D.keys():
        bname = D[key]
        b = arm.data.bones[bname]
        mat_local = b.matrix_local.transposed()
#        print('\nb.name',b.name)
#        print('b.parent.name = ',b.parent)
#        print(mat_local)
        m = mat_local
        mat = Matrix([[m[0][0],m[0][1],m[0][2]],
                     [m[1][0],m[1][1],m[1][2]],
                     [m[2][0],m[2][1],m[2][2]]]).transposed()
#        print('mat = ')
#        print(mat)
        loc = [mat_local[3][0],mat_local[3][1],mat_local[3][2]]
        quat = Matrix(mat).transposed().to_quaternion()   
#        print('quat =  ',quat)
#        print('loc = ',loc)               
#        print('key : ',key)
        bns += [b.name]
        bplist[b.name]= {}
        root_bone = False
#        if b.name == 'root':
        if b.parent == None:
            root_bone = True
            bplist[b.name]['parent'] = 'world'
        else:
            bparent = b.parent
#            print('bparent = ',bparent.name)
            parent_index = bns.index(bparent.name)
#            print('parent_index = ',parent_index)
            bplist[b.name]['parent'] = bns[parent_index]
        bplist[b.name]['quat'] = quat
        bplist[b.name]['mat'] = mat
        bplist[b.name]['location'] = loc
#
# Data generated
# compute qbrs 
#        print('bns = ',bns)
    bn  = 0
    rel = Vector([0,0,0])
    while bn < len(bns):
        root_bone = False
        if bn == 0:
            root_bone = True
        name = bns[bn]
        q = bplist[name]['quat']
        mat = bplist[name]['mat']
        loc = bplist[name]['location']
#        if name == 'root':
        if root_bone:
            qref = mr.to_quaternion().conjugated()
            locp = Vector([0,0,0])
        else:
            parent = bplist[name]['parent']
            qref = bplist[parent]['quat'].conjugated()
            locp = bplist[parent]['location']
#    qstar = q.conjugated()
        qstar = q
        qbr = quat_mult(qref,qstar)
        for i in range(3):
            rel[i] = loc[i] - locp[i]
#        if name == 'root':
        if root_bone:
            pos = Vector([rel[0],rel[1],rel[2]])
        else:
            mparent = mrt*qref.conjugated().to_matrix()
            pos = mparent*Vector([rel[0],rel[1],rel[2]])
#        print('\n',name)
#        print('bqr = %10.7f %10.7f %10.7f %10.7f'%(qbr[0],qbr[1],qbr[2],qbr[3]))
#        print('    locs = %10.7f %10.7f %10.7f'%(pos[0],pos[1],pos[2]))
        bqr += [qbr]
        locs += [pos]
        bn += 1
def tm_to_blender(x,y,z):
    seq = 'YZX'
    angles = (math.radians(z),math.radians(x),math.radians(y))
    ea = mathutils.Euler(angles,seq)
    quat = ea.to_quaternion()
    return quat
def get_empty(name_in):
#    print('Looking for',name)
#    print(' len(name) = ',len(name.rstrip("\x00")))
    name = name_in.rstrip("\x00")
    scn = bpy.context.scene
    objects = scn.objects
    for obj in objects:
        if obj.type == 'EMPTY':
#            print('{0:52s} {1:15s} '.format(obj.name,obj.type))
            if 'grpName' in obj:
                grpName = obj['grpName']
                if grpName == name:
                    print(' found match ')
                    return obj
    obj = bpy.ops.object.empty_add(type='PLAIN_AXES',location=(0,0,0))
    bpy.data.objects['Empty'].name = name
    print('found no match')
    return None
def write_vec(lbl,v):
    fd.write(lbl+'%9.6f %9.6f %9.6f\n'%(v[0],v[1],v[2])) 
def write_mat(lbl,m):
    fd.write(lbl+'\n')
    for i in range(4):
        fd.write('%9.6f %9.6f %9.6f %9.6f'%(m[i][0],
            m[i][1],m[i][2],m[i][3])+'\n')  
def load_anm (filename):
    global origins,compression,arms,animorigins,bone_anims,prop_anims
    global animorigins,fd
    scene = bpy.context.scene
    origins = []
    compression = []
    start_time = time.clock()
    prev_time = start_time
#    in_editmode = Window.EditMode()
#    if in_editmode:
#        Window.EditMode(0)
    stream = open(filename,"rb")
    filepath = os.path.dirname(filename)
    filebase = os.path.basename(filename)
    fullbase = os.path.splitext(filebase)[0]
    basename = fullbase[0:21] #Blender has a $%$@ 21-char limit!
    basedir = os.path.split(filepath)[0]
#    scene = Blender.Scene.GetCurrent()
#scene.setLayers([1])

#Start Reading!
    header = anm_header(stream)
    if header.Magic != 12:
        raise IOError("File is not a TM Animation")
#Get origin matrices for each subsequent animation -- first must use OriginInWorld?
    print ('header processed: {0:.4f} sec.'.format(time.clock()-prev_time))
    prev_time = time.clock()

    framecount = header.count_frames
    animorigins = []
    animorigins.append(header.matrix)
    for a in range(header.count_anims-1):
        animorigins.append(anm_origin(stream).origin)
#Read bone-based anim data
    bone_anims = {}
    for ba in range(header.count_anims):
        bone_anims[ba] = BoneAnims(stream)
# save destination in scene properties
        hdr = bone_anims[ba].header
        bakey = 'Frame Count'
        if (bakey in scene) == False:
            scene[bakey] = framecount
        else:
            scene[bakey] = framecount
        dest = (hdr.dest_x,hdr.dest_y,hdr.dest_z,hdr.dest_angle)
        print('dest = ',dest)
        bakey = 'BA_DEST'+str(ba)
        print('ba key = ',bakey)
        if (bakey in scene) == False:
            scene[bakey] = {}
        scene[bakey] = dest
# save origin in scene properties
        bakey = 'BA_ROTATION'+str(ba)
        if (bakey in scene) == False:
            scene[bakey] = {}
        scene[bakey] = animorigins[ba].to_euler()
        bakey = 'BA_OFFSET' + str(ba)
        if (bakey in scene) == False:
            scene[bakey] = {}
        scene[bakey] = [animorigins[ba][3][0],
                        animorigins[ba][3][1],
                        animorigins[ba][3][2]]
#Read prop anim data
    print ('{0:d} Bone Anims processed: {1:.4f} sec.'.format(header.count_anims, (time.clock()-prev_time)))
    prev_time = time.clock()
    prop_anims = {}
    for pa in range(header.count_static_anims):
        tprop = PropAnim(stream, framecount)
        prop_anims[tprop.name] = tprop
        print('tprop.name = ',tprop.name)
    print('{0:d} Prop Anims processed: {1:.4f} sec.'.format(header.count_static_anims, (time.clock()-prev_time)))
    prev_time = time.clock()
    print ('basename = ',basename)
    print ('framecount = ',framecount)
    print('{0:10}: {1:d} Frames Imported: {2:.4f} sec.'.format(basename, framecount, (time.clock() - start_time)))
#Now do something with it! Translate into Blender stuff

    
#    scene.frame_start = 1
    scene.frame_end = scene.frame_start + framecount - 1
#Check for armatures
    obs = []
    armatures = []
    bas = []
    selected_objects = bpy.context.selected_objects
    active_obj = bpy.context.scene.objects.active
    arms = []
    if active_obj.type != 'ARMATURE':
        for ob in bpy.data.objects:
            if ob.type == 'ARMATURE':
                if 'BoneOrder' in ob:
                    nab = len(ob['BoneOrder'].values())
                else:
                    nab = len(ob['TheMovies']['BoneOrder'].values())
                armatures += [[ob,nab]]
        for ba in bone_anims:
            nbb = bone_anims[ba].header.bone_count
            for armature in armatures:
                if armature[1] == nbb:
                    arms +=[armature[0]]
                    armatures.remove(armature)
                    break
    else:
        arms += [active_obj]
        for obj in selected_objects:
            if obj.type == 'ARMATURE':
                if (obj in arms) == False:
                    arms += [obj] 
        
    for arm in arms:
        print('arm ',arm.name)
# active object
# bpy.context.scene.objects.active                
    print(' BA ARMATURE')
#    fd = open('C:\TMBlender\Ble279\ScriptDevelopment1\pms.txt','w')
    ba_index = 0
    print('bone_anims.keys = ',bone_anims.keys())
    for ba in bone_anims.keys():
        print('ba = ',ba)
#        print('{0:3d} {1:s}'.format(ba,obs[ba].name))
        kf = compression[ba]
        nbones = bone_anims[ba].header.bone_count
        print('nbones = ',nbones)
#        bonelist.bone_name_list(nbones)       
        btag = 0
        rs = 180.0
        arm_ob = arms[ba_index]
        if('BAOBJ' in arm_ob) == False:
            arm_ob['BAOBJ'] = {}
        arm_ob['BAOBJ'] = ba
        generate_qbrs(arm_ob)
        print('ba = ',ba,' arm_ob = ',arm_ob.name)
        nbb = bone_anims[ba].header.bone_count
        print('ba ',ba,' has ',nbb,' bones')
        if 'BoneOrder' in arm_ob:
            nab = len(arm_ob['BoneOrder'].values())
        else:
            nab = len(arm_ob['TheMovies']['BoneOrder'])
        print('armature ',arm_ob.name,' has ',nab,'bones.')
        arm_ob.select = True
        bpy.context.scene.objects.active = arm_ob
        print('context.area.type = ',bpy.context.area.type)
        print('context.mode      = ',bpy.context.mode)
#        bpy.ops.object.mode_set(mode='POSE')

#        bpy.ops.pose.rot_clear()
#        bpy.ops.pose.loc_clear()
#        bonelist.bone_name_list(nbones)
#        bns = bonelist.bns
#        locs = bonelist.locs
#        bqr = bonelist.bqr
#        fd.write('arm_ob: %s\n'%arm_ob.name)
        ba_index += 1
        for bi,ba_obj in bone_anims[ba]:
#            print(btag,bi,bns[btag])
            bs = bqr[btag]
            bq = mathutils.Quaternion(bs)
#            print('framecount = ',framecount)
            bname = bns[btag]
#            pbone = bpy.data.objects['507812352'].pose.bones[bname]
            pbone = arm_ob.pose.bones[bname]            
            for f in range(framecount):
                scene.frame_current = scene.frame_start + f
                (u,v,w) = (rs*ba_obj.x[f],rs*ba_obj.y[f],rs*ba_obj.z[f])
#                print(u,v,w)
                x = ba_obj.u[f]*kf - locs[btag][0]
                y = ba_obj.v[f]*kf - locs[btag][1]
                z = ba_obj.w[f]*kf - locs[btag][2]
                rel = Vector((0,0,0))
                pbone.rotation_quaternion = mathutils.Quaternion([1,0,0,0])
                pm = pbone.matrix.transposed()
                if btag == 0:
#                    if f == 0:
#                        write_mat('pm ',pm)
#                    rel = Vector((-y,z,-x))
                    rel = pm*Vector((x,y,z))
                elif pbone.name == 'control2_arm_r' or pbone.name == 'control2_arm_l':
                    rel = Vector((x,-y,-z))
                quat = tm_to_blender(u,v,w)
                pbone.rotation_quaternion = quat_mult(quat,bq)
                pbone.location = rel
                pbone.keyframe_insert('rotation_quaternion')
                pbone.keyframe_insert('location')
            btag += 1
    for pa in prop_anims.keys():
        prop_anim = prop_anims[pa]
        prop_obj = get_empty(pa)
        if prop_obj != None:
            print('prop_obj = ',prop_obj.name,' pa = ',pa)
            comp = prop_anim.header.compression
            for f in range(framecount):
                scene.frame_current = scene.frame_start + f
#                Blender.Set('curframe',f+f0)
                frame = prop_anim.frames[f]
                (u,v,w) = (math.pi*frame.u,math.pi*frame.v,math.pi*frame.w)
                
                (x,y,z) = (comp*frame.x,comp*frame.y,comp*frame.z)
#               print('uvw = ',u,v,w)
                prop_obj.location = Vector((x,y,z))
                prop_obj.rotation_euler = mathutils.Euler((u,v,w))
                prop_obj.keyframe_insert('rotation_euler')
                prop_obj.keyframe_insert('location')
#                prop_ob.setEuler(u,v,w)
#                prop_ob.setLocation(x,y,z)
#                prop_ob.insertIpoKey(Blender.Object.ROT)
#                prop_ob.insertIpoKey(Blender.Object.LOC)            
#    fd.close()        
#    apply_move()


def anm_in():
    bpy.ops.object.animation_import('INVOKE_DEFAULT')
if __name__ == '__main__':
    register()


