#!BPY

"""
Name: 'The Movies (.msh) 2009...'
Blender: 279
Group: 'Export'
Tooltip: 'Export to Lionhead Studios The Movies file format. (.msh)'
"""

__author__ = ["Glen Rickey, Nick Hudson, Mark S Andrews"]
__url__ = ("Director's Cut Modding Foundry","http://www.dcmodding.com")
__version__ = "1.04 05-07-2009"
__bpydoc__ = """\

Msh Exporter

This script exports a msh file for use with Lionhead Studio Limited's
The Movies. It is not supported by either Lionhead or Activision.

It does not currently support all features of the msh files.

Known Problems:

ConvexHulls are not exported.
Requires very properly formatted file.  Use tm_preflight.py script to condition scene data first.
Mod for blender 2.79b rev 3

Converted to Blender 2.79b by Cary Semar  1/1/2022
                                Released: 1/3/2022

Validation:
  import with 249b.  Open blend in 279b and
  export with tm_msh_export3.  Works in game.
  2/2/2022.  Double verified.

Validation:
    import with tm_msh_import3, export to game with tm_msh_export3.
    gorilla          OK
    fut_robot_tz2    OK
    horse_v00        OK
    p_gun_ak47       OK   firing effect checks out.
    cos_f_scientist

    The use_alpha flag needs to be set to on by import script.  This
    script uses that to configure materials for output.
    
  
"""

# ***** BEGIN GPL LICENSE BLOCK *****
#
# Script copyright (C) Mark S Andrews, Nick Hudson, Glen Rickey
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****
# --------------------------------------------------------------------------
bl_info = {"name": "The Movies Mesh Export", 
    "description": "Exports models to The Movies",
    "category": "Import-Export",
    "blender": (2,79,2),
    "location": "File > Export > The Movies Mesh (.msh)",
    "author": "Nick Hudson, Glen Rickey, Mark S Andrews, Cary Semar"
}
# Import modules

#import Blender
#from Blender import Image, Material, Texture, NMesh, Window, Armature, Group
#from Blender.Mathutils import Matrix, Vector
from random import randint
import struct, os, re
import tmConst
import bpy, os, struct, re, math, bmesh
from time import time
from bpy.types import Operator
from bpy.props import StringProperty, BoolProperty, EnumProperty
from bpy_extras.io_utils import ExportHelper
import pdb
print('Installed: tm_msh_export ver 3.4')
#***********************************************
# globals
#***********************************************

material_list = []
material_keys = []

unnamed_group = re.compile(r"(\d{2})(\.\d+)?")
static_anim_group = re.compile(r".?_[sc]a_")
#grouped_mesh = re.compile(r'(.+)\.(\d\d)\..+')
#numbered_mesh = re.compile(r'(.+)\.\d+')
#room_mesh = re.compile(r'(.+)\.room\_')


#***********************************************
# helper classes & functions for msh file
#***********************************************

def write_nts(item, stream, length=32):
    if len(item) > length:
        raise ValueError("%s is longer than %d"%(item, length))
    while len(item) < length:
#        item += bytes([0])
        item += "\x00"
#    data = bytes(str(data).encode("utf-8"))
    stream.write(bytes(str(item).encode("utf-8")))

def write_null(struct_type, stream):
    data = ""
    for b in range(struct.calcsize(struct_type)):
        data += "\x00"
    stream.write(data)


class range_adjust(object):
    def __init__(self, min, max):
        self.min = min
        self.max = max
        self.range = abs(self.max - self.min)
    def adjust(self, value):
        if self.range == 0:
            return 0
        scale = 65535.0 / self.range
        return int((value - self.min) * scale)
    
def getNeckIndex(point, snap_points):
        '''
        Returns the closest vec to snap_points
        '''
        close_dist= 1<<30
        close_vec= None

        x= point[0]
        y= point[1]
        z= point[2]
        for v in snap_points:
                # quick length cmp before a full length comparison.
                if abs(x-v[0]) < close_dist and\
                abs(y-v[1]) < close_dist and\
                abs(z-v[2]) < close_dist:
                        l= (v-point).length
                        if l<close_dist:
                                close_dist= l
                                close_vec= v
        return snap_points.index(close_vec)
    
def adjustBounds(obj):
    mat = obj.matrix_world
    if obj.type == 'MESH':
#        msh = obj.getData(0,1)
        msh = obj.data
        worldverts = [(x.co * mat) for x in msh.vertices]
        xMin = min([x[0] for x in worldverts])
        xMax = max([x[0] for x in worldverts])
        yMin = min([y[1] for y in worldverts])
        yMax = max([y[1] for y in worldverts])
        zMin = min([z[2] for z in worldverts])
        zMax = max([z[2] for z in worldverts])
        
        return [
                min([xMin, Bounds[0]]),
                max([xMax, Bounds[1]]), 
                min([yMin, Bounds[2]]),
                max([yMax, Bounds[3]]), 
                min([zMin, Bounds[4]]),
                max([zMax, Bounds[5]])
                ]
        
    elif obj.type == 'EMPTY':
#        loc = obj.getLocation('worldspace')
        loc = obj.location
        return [
                min([loc[0], Bounds[0]]),
                max([loc[0], Bounds[1]]), 
                min([loc[1], Bounds[2]]),
                max([loc[1], Bounds[3]]), 
                min([loc[2], Bounds[4]]),
                max([loc[2], Bounds[5]])
                ]
    
def calc_bbox(bounds_list):
    # Compute new extents of bounding box
    xMin = min(map(lambda x: x[0], bounds_list))
    xMax = max(map(lambda x: x[0], bounds_list))
    yMin = min(map(lambda y: y[1], bounds_list))
    yMax = max(map(lambda y: y[1], bounds_list))
    zMin = min(map(lambda z: z[2], bounds_list))
    zMax = max(map(lambda z: z[2], bounds_list))

    return [xMin, xMax, yMin, yMax, zMin, zMax] 


class triface(object):
    def __init__(self, vindex=(0,0,0), faceuvs=None, lmuvs=None):
        self.vertex_index = vindex
        self.faceuvs = faceuvs
        self.lmuvs = lmuvs
    
def getVertsFromGroup(grp,ob):
    vertices = []
    e_idx = ob.vertex_groups[grp].index
    vs = [v for v in ob.data.vertices if (v.groups)]
    for v in vs:
        for g in v.groups:
            if (g.group == e_idx):
                vertices.append(v)
                break
    return vertices
#***********************************************
# writer functions for msh file
#***********************************************

def write_material(mat, stream):
    image_list = bpy.data.images.keys()
    texture_list = bpy.data.textures
#    print('image_list = ',image_list)
    i = [-1,-1,-1,-1]
    use_alpha = False
    for slot in range(4):
        texture_slot = mat.texture_slots[slot]
#        print('dir texture_slot')
#        print(dir(texture_slot))
#        print(ksx)
        if slot == 0:
            use_alpha = texture_slot.use_map_alpha
        for texture in texture_list:
            if texture.type == 'IMAGE':
                if texture.image != None:
                    img = texture.image.name
                    if texture_slot.texture.type == 'IMAGE':
                        if texture_slot.texture != None:
                            if texture_slot.texture != None:
                                img = texture_slot.texture.image
                                if img != None:
                                    i[slot] = image_list.index(img.name)
   
    # *********** write image refs *************
    stream.write(struct.pack("bbbb", i[0], i[1], i[2], i[3]))
    #  Material Flags
    matflags0 = 0
    matflags1 = 4 #mesh always has UV
    matflaglist = [1,1,0,0,0,0,0,0,0]
    matscrollU = 0
    matscrollV = 0
    matrotUV = 0
#    if mat.properties.has_key('TheMovies'):
    if 'TheMovies' in mat:
        if 'doublesided' in mat:
            matflags0 |= ((mat['doublesided'] % 2) * 1)
            matflags0 |= ((mat['floor_shadow_tex'] % 2) * 2)
            matflags0 |= ((mat['alpha_separate'] % 2)* 4)
        
            matflaglist[0] = mat['wrap_U'] % 2
            matflaglist[1] = mat['wrap_V'] % 2
            matflaglist[2] = use_alpha
            matflaglist[3] = mat['enable_alpha_test'] % 2
            matflaglist[4] = mat['glass'] % 2
            matflaglist[5] = mat['water'] % 2
            matflaglist[6] = mat['still_water'] % 2
            matflaglist[7] = mat['alpha_per_vertex'] % 2
            matflaglist[8] = mat['alphaenvmap'] % 256
        
            matflags1 |= ((mat['invisible'] % 2) * 1)
            matflags1 |= ((mat['not_z_write'] % 2) * 2)
            matflags1 |= ((mat['self_lit'] % 2) * 8)
            matflags1 |= ((mat['no_floor_shadow'] % 2) * 16)
            matflags1 |= ((mat['no_delaydraw_transp'] % 2) * 32)
            matflags1 |= ((mat['tri_sort'] % 2) * 64)
            matflags1 |= ((mat['additive'] % 2) * 128)

            matscrollU = mat['scroll_U'] % 256
            matscrollV = mat['scroll_V'] % 256
            matrotUV = mat['rot_UV'] % 256
        else:
            matflags0 |= ((mat['TheMovies']['doublesided'] % 2) * 1)
            matflags0 |= ((mat['TheMovies']['floor_shadow_tex'] % 2) * 2)
            matflags0 |= ((mat['TheMovies']['alpha_separate'] % 2)* 4)
        
            matflaglist[0] = mat['TheMovies']['wrap_U'] % 2
            matflaglist[1] = mat['TheMovies']['wrap_V'] % 2
            matflaglist[2] = use_alpha
            matflaglist[3] = mat['TheMovies']['enable_alpha_test'] % 2
            matflaglist[4] = mat['TheMovies']['glass'] % 2
            matflaglist[5] = mat['TheMovies']['water'] % 2
            matflaglist[6] = mat['TheMovies']['still_water'] % 2
            matflaglist[7] = mat['TheMovies']['alpha_per_vertex'] % 2
            matflaglist[8] = mat['TheMovies']['alphaenvmap'] % 256
        
            matflags1 |= ((mat['TheMovies']['invisible'] % 2) * 1)
            matflags1 |= ((mat['TheMovies']['not_z_write'] % 2) * 2)
            matflags1 |= ((mat['TheMovies']['self_lit'] % 2) * 8)
            matflags1 |= ((mat['TheMovies']['no_floor_shadow'] % 2) * 16)
            matflags1 |= ((mat['TheMovies']['no_delaydraw_transp'] % 2) * 32)
            matflags1 |= ((mat['TheMovies']['tri_sort'] % 2) * 64)
            matflags1 |= ((mat['TheMovies']['additive'] % 2) * 128)

            matscrollU = mat['TheMovies']['scroll_U'] % 256
            matscrollV = mat['TheMovies']['scroll_V'] % 256
            matrotUV = mat['TheMovies']['rot_UV'] % 256    
    # write Material Flags
    stream.write(struct.pack("B",matflags0))
    for flag in matflaglist:
        stream.write(struct.pack("B", flag))
    stream.write(struct.pack("B", matflags1))
#    stream.write("\00") #padding
    stream.write(struct.pack("B",0))
    # write material color
    (B,G,R) = mat.diffuse_color
    stream.write(struct.pack("BBBB", int(B * 255), int(G * 255), int(R * 255), int(mat.alpha * 255)))
    # write material scroll properties
    stream.write(struct.pack("BBB", matscrollU, matscrollV, matrotUV))
#    stream.write("\x00") #padding
    stream.write(struct.pack("B",0))

#***********************************************
# Write the control meshes in the correct order.
#***********************************************
def write_control_meshes(stream):
    '''This sets the order for writing out a control mesh, since
       we are not sure what order these meshes may come in'''
    if clickable is not None:
        write_basic_mesh(clickable, stream)

    if collision is not None:
        write_basic_mesh(collision, stream)

    if shadow is not None:
        write_basic_mesh(shadow, stream)

    if z_height is not None:
        write_basic_mesh(z_height, stream)

    if neg_space is not None:
        write_basic_mesh(neg_space, stream)

    if min_outline is not None:
        write_basic_mesh(min_outline, stream)

    if lot_boundary is not None:
        write_basic_mesh(lot_boundary, stream)



def write_basic_mesh(mesh, stream):
    '''This is intended for writing out control meshes only, which
       require less data.'''

    meshdata = mesh.getData()
    
    # ********** write vertex count ************
    stream.write(struct.pack("L",len(meshdata.verts)))
    # stream.write(len(mesh.verts))
    # ********** write face count **************
    stream.write(struct.pack("L",len(meshdata.faces)))
    #stream.write(len(mesh.faces))
    for v in meshdata.verts:
        # ******* write vertex position ********
        stream.write(struct.pack("fff", v.co[0], v.co[1], v.co[2]))
    for f in meshdata.faces:
        v0 = meshdata.verts.index(f.v[0])
        v1 = meshdata.verts.index(f.v[1])
        v2 = meshdata.verts.index(f.v[2])
        # ******** write face vertices *********
        stream.write(struct.pack("HHH", v0, v1, v2))
    if len(meshdata.faces) & 1:
        # ********** write pad bytes ***********
        stream.write("\x00\x00")

def write_group(grp, stream):
    print('writing group = ',grp.keys())
    grp_keys = grp.keys()
    grp_keys.sort()
    meshes = []
    grpflags1 = 0
    grpflags2 = 0
    has_hidden = 0
    hidden_on = 0
    hidden_off = 0
#    pdb.set_trace()
    if len(grp) == 0:
        return
    print('grp = ',grp)
#    for ob in grp:
    for key in grp_keys:
        ob = grp[key]
        print('ob = ',ob.name,' ob.type = ',ob.type)
        if ob.type == "EMPTY":
            # get pivot Matrix
            m = ob.matrix_world
            pivot = [[m[0][0],m[1][0],m[2][0],m[3][0]],
            [m[0][1],m[1][1],m[2][1],m[3][1]],
            [m[0][2],m[1][2],m[2][2],m[3][2]],
            [m[0][3],m[1][3],m[2][3],m[3][3]]]
            #get ID Properties
#            if ob.properties.has_key('TheMovies')==True:
            if 'TheMovies' in ob:
                if 'has_transanim' in ob:
                    #flags1
                    grpflags1 |= ((ob['has_transanim'] % 2))
                    grpflags1 |= ((ob['is_a_carbody'] % 2) * 2)
                    #flags2
                    grpflags2 |= ((ob['is_land'] % 2) * 4)
                    grpflags2 |= ((ob['hide_reflection'] % 2) * 8)
                    has_hidden = ob['has_hidden'] % 2
                    hidden_on = ob['hidden_on']
                    hidden_off = ob['hidden_off']
                    grpflags2 |= (has_hidden * 16)
                else:
                    #flags1
                    grpflags1 |= ((ob['TheMovies']['has_transanim'] % 2))
                    grpflags1 |= ((ob['TheMovies']['is_a_carbody'] % 2) * 2)
                    #flags2
                    grpflags2 |= ((ob['TheMovies']['is_land'] % 2) * 4)
                    grpflags2 |= ((ob['TheMovies']['hide_reflection'] % 2) * 8)
                    has_hidden = ob['TheMovies']['has_hidden'] % 2
                    hidden_on = ob['TheMovies']['hidden_on']
                    hidden_off = ob['TheMovies']['hidden_off']
                    grpflags2 |= (has_hidden * 16)
        if ob.type == "MESH":
            meshes.append(ob)
    print('pivot = ',pivot)    
    # *********** write group header ***********
    stream.write(struct.pack("BBBB", len(meshes), 0, grpflags1, grpflags2))
    # ************** write matrix **************
    for p in range(4):
        stream.write(struct.pack("fff", pivot[p][0], pivot[p][1], pivot[p][2]))
    
    if has_hidden:
        stream.write(struct.pack("HH", hidden_on, hidden_off))
        
#    for m in reversed(meshes):
    for m in meshes:
        print(' writing mesh = ',m.name)
        write_mesh(m, stream)

def write_mesh(ob, stream):
    # this is the main function for writing a mesh to file. This is for
    # normal meshes only, and should not be used for control meshes 
    # (which don't have textures anyway

    global material_keys
    global bone_list
    print('\nwriting mesh ',ob.name,' at ',stream.tell())
    # Get ID Properties
    meshflags1 = 32 # all meshes are compressed
    meshflags2 = 0
    is_boned = 0
    has_meshid = 1
    generate_new_id = 1
    has_neckconnect = 0 # we're going to stick to the format, even if Lionhead didn't...
    neckconnect_count = 0
    has_lightmap = 0
    bone_per_vertex = 0
    unkflag = 0
    indexid = 0
    vertexid = -1
    skeletonid = 0
    #mesh = ob.getData()
    #mobj = ob.getData(False,True)
#    if ob.mode == 'OBJECT':
#        bpy.ops.object.editmode_toggle()
    mesh = ob.data
    mobj = ob.data
    
    #if 'LightMap' in mobj.getUVLayerNames():
    #    has_lightmap = 1
    if len(ob.vertex_groups) > 0:
        is_boned = 1
        meshflags1 |= 1

#    if mesh.properties.has_key('TheMovies')==True:
#    print('ID Property: ',mesh['has_neckconnect'])
    if 'TheMovies' in mesh:
        if 'has_floor_reflections' in mesh:
            #meshflags1
            meshflags1 |= ((mesh['has_floor_reflections'] % 2) * 2)
            meshflags1 |= ((mesh['no_outline'] % 2) * 4)
            meshflags1 |= ((mesh['is_landscape'] % 2) * 8)
            has_meshid = mesh['has_meshid'] % 2
            generate_new_id = mesh['generate_new_id'] % 2
            has_neckconnect = mesh['has_neckconnect'] % 2
            meshflags1 |= ((mesh['accepts_actor_shadow'] % 2) * 128)
            #meshflags2
            meshflags2 |= has_lightmap
            meshflags2 |= ((mesh['is_minutehand'] % 2) * 2)
            meshflags2 |= ((mesh['is_hourhand'] % 2) * 4)
            meshflags2 |= ((mesh['static_backdrop'] % 2) * 8)
            unkflag = mesh['unk_flag']
            bone_per_vertex = mesh['bone_per_vertex']
            if has_meshid:
                if generate_new_id:
                        indexid = randint(0,4294967295)
                        vertexid = -1
                        if is_boned:
                            skeletonid = randint(0,4294967295)
                else:
                    indexid = mesh['indexid']
                    vertexid = mesh['vertexid']
                    skeletonid = mesh['skeletonid']
        else:
            #meshflags1
            meshflags1 |= ((mesh['TheMovies']['has_floor_reflections'] % 2) * 2)
            meshflags1 |= ((mesh['TheMovies']['no_outline'] % 2) * 4)
            meshflags1 |= ((mesh['TheMovies']['is_landscape'] % 2) * 8)
            has_meshid = mesh['TheMovies']['has_meshid'] % 2
            generate_new_id = mesh['TheMovies']['generate_new_id'] % 2
            has_neckconnect = mesh['TheMovies']['has_neckconnect'] % 2
            meshflags1 |= ((mesh['TheMovies']['accepts_actor_shadow'] % 2) * 128)
            #meshflags2
            meshflags2 |= has_lightmap
            meshflags2 |= ((mesh['TheMovies']['is_minutehand'] % 2) * 2)
            meshflags2 |= ((mesh['TheMovies']['is_hourhand'] % 2) * 4)
            meshflags2 |= ((mesh['TheMovies']['static_backdrop'] % 2) * 8)
            unkflag = mesh['TheMovies']['unk_flag']
            bone_per_vertex = mesh['TheMovies']['bone_per_vertex']
            if has_meshid:
                if generate_new_id:
                        indexid = randint(0,4294967295)
                        vertexid = -1
                        if is_boned:
                            skeletonid = randint(0,4294967295)
                else:
                    indexid = mesh['TheMovies']['indexid']
                    vertexid = mesh['TheMovies']['vertexid']
                    skeletonid = mesh['TheMovies']['skeletonid']
    print('447: has_neckconnect = ',has_neckconnect)       
#    if 'neckconnect' in mesh.getVertGroupNames():
    print('group keys = ',ob.vertex_groups.keys())
    print('neckconnect_count = ',neckconnect_count)
    if 'neckconnect' in ob.vertex_groups.keys():
        has_neckconnect=1
        neckconnect_count = len(getVertsFromGroup('neckconnect',ob))
    print('neckconnect_count = ',neckconnect_count)
                
    meshflags1 |= has_meshid * 16
    meshflags1 |= has_neckconnect * 64
    print('mobj = ',mobj.name)
    mat = mobj.materials[0]
    print('mat = ',mat.name)
    # Start Writin' !!
    # ********** write material index **********
    print('writing material index')
    print('start writing: ',stream.tell())
    print('material_keys = ',material_keys)
    print('material_index = ',material_keys.index(material_keys[0]))
    if mat:
        stream.write(struct.pack("L",material_keys.index(mat.name)))
    #Get all faces
    face_list = []
#    has_uv = mobj.faceUV
    has_uv = True
    if mesh.uv_layers == None:
        has_uv = False
    
    if not has_uv:
        face_uv = None
    for f in ob.data.polygons:
        if len(f.vertices)==3:
            f_verts = []
            for vert_idx in f.vertices:
                f_verts += [vert_idx]
                f_uv = []
                for loop_idx in f.loop_indices:
                    f_uv += [(ob.data.uv_layers.active.data[loop_idx].uv)]
            new_face = triface((f_verts[0],f_verts[1],f_verts[2]))
            new_face.faceuvs = f_uv[0],f_uv[1],f_uv[2]
#            print('new_face = ',new_face)
#            if (has_uv): 
#                mobj.activeUVLayer = 'UVTex'
#                f_uv = face.
#                f_uv = (mesh.vertices[iv1.index],mesh.vertices[iv2.index],mesh.vertices[iv2.index])
#                new_face.faceuvs = f_uv[0], f_uv[1],f_uv[2]
            if (has_lightmap):
                mobj.activeUVLayer = 'LightMap'
                f_uv = face.uv
                new_face.lmuvs = f_uv[0], f_uv[1], f_uv[2]
            face_list.append(new_face)
        
        else: #face is a quad, triangulate it!
            new_face = triface((f_v[0].index, f_v[1].index, f_v[2].index))
            new_face_2 = triface((f_v[0].index, f_v[2].index, f_v[3].index))
            
            if (has_uv):
                mobj.activeUVLayer = 'UVTex'
                f_uv = face.uv
                new_face.faceuvs = f_uv[0], f_uv[1], f_uv[2]
                new_face_2.faceuvs = f_uv[0], f_uv[2], f_uv[3]
            if (has_lightmap):
                mobj.activeUVLayer = 'LightMap'
                f_uv = face.uv
                new_face.lmuvs = f_uv[0], f_uv[1], f_uv[2]
                new_face_2.lmuvs = f_uv[0], f_uv[2], f_uv[3]
            
            face_list.append( new_face )
            face_list.append( new_face_2 )
    print('len(face_list) = ',len(face_list) )       
    #Process Faces
    vlist = {}
    uvlist  =  {}
    lmuvlist = {}
    dupverts = {}
    new_index = 0
    index_list = []
    
    #calc ranges 

    tx_min = min([f.faceuvs[i][0] for f in face_list for i in range(3)])
    tx_max = max([f.faceuvs[i][0] for f in face_list for i in range(3)])
    ty_min = min([f.faceuvs[i][1] for f in face_list for i in range(3)])
    ty_max = max([f.faceuvs[i][1] for f in face_list for i in range(3)])
    txr = range_adjust(tx_min, tx_max)    
    tyr = range_adjust(ty_min, ty_max)
    if has_lightmap:
        lmx_min = min([f.lmuvs[i][0] for f in face_list for i in range(3)])
        lmx_max = max([f.lmuvs[i][0] for f in face_list for i in range(3)])
        lmy_min = min([f.lmuvs[i][1] for f in face_list for i in range(3)])
        lmy_max = max([f.lmuvs[i][1] for f in face_list for i in range(3)])
        lmxr = range_adjust(lmx_min, lmx_max)
        lmyr = range_adjust(lmy_min, lmy_max)
    else:
        lmxr = range_adjust(0,0)
        lmyr = range_adjust(0,0) 
    xr = range_adjust(Bounds[0], Bounds[1])
    yr = range_adjust(Bounds[2], Bounds[3])
    zr = range_adjust(Bounds[4], Bounds[5])
    nr = range_adjust(-1.0, 1.0)
    
    # first add all the valid verts (those in actual faces)
    goodindx = []
#    for tf in mobj.faces:
#        for gv in tf.v:
#            if gv.index not in goodindx:
#                goodindx.append(gv.index)
    for tf in mobj.polygons:
        for v in range(len(tf.vertices)):
            vtx = tf.vertices[v]
            if vtx not in goodindx:
                goodindx.append(vtx)
    goodindx.sort()         
    for vtex in mobj.vertices:
        if vtex.index in goodindx:
            #multiply by objects' 4x4 matrix to get World Coordinates...
            vworld = vtex.co * ob.matrix_world   # or ob.matrix_local ?
            vlist[vtex.index] = [xr.adjust(vworld[0]), \
                yr.adjust(vworld[1]), \
                zr.adjust(vworld[2]), \
                nr.adjust(vtex.normal[0]), \
                nr.adjust(vtex.normal[1]), \
                nr.adjust(vtex.normal[2])]
            #populate the uv / lm lists with spaces
            uvlist[vtex.index] = []
            dupverts[vtex.index] = []
            if (has_lightmap):
                lmuvlist[vtex.index] = []
        
    new_index = max(vlist)+1 #AHA!!
    
    for f in face_list:
        for i in range(3):

            uv = [txr.adjust(f.faceuvs[i][0]),tyr.adjust(f.faceuvs[i][1])]
            v_indx = f.vertex_index[i]
            if (has_lightmap):
                lmuv = [lmxr.adjust(f.lmuvs[i][0]),lmyr.adjust(f.lmuvs[i][1])]
                if uvlist[v_indx] != []:
                    rootindx = dupverts[v_indx][0]
                    matches = []
                    for ckvert in dupverts[v_indx]:
                        if uvlist[ckvert] == uv:
                            matches += [ckvert]
                    if len(matches) == 0:
                    #must be a new UV!
                        #duplicate existing vertex
                        vlist[new_index] = vlist[rootindx]
                        #add its new uv
                        uvlist[new_index] = uv
                        lmuvlist[new_index] = lmuv
                        #add this index to a list so we can find it later
                        dupverts[rootindx] += [new_index]
                        v_indx = new_index
                        new_index += 1
                    else:
                        #Check our lmuvs
                        if lmuvlist[v_indx] != []:
                            matched = False
                            for ckvert in matches:
                                if lmuvlist[ckvert] == lmuv:
                                    matched = True
                                    v_indx = ckvert
                            if matched == False:
                            #Must be a new LMUV!
                                #duplicate this vert/uv
                                vlist[new_index] = vlist[rootindx]
                                #add its uv
                                uvlist[new_index] = uv
                                lmuvlist[new_index] = lmuv
                                #add this index to a list so we can find it later
                                dupverts[rootindx] += [new_index]
                                v_indx = new_index
                                new_index += 1
                        else:
                            #first time we've seen this.  Add it.
                            lmuvlist[v_indx] = lmuv
                        
                else:
                    uvlist[v_indx] = uv
                    dupverts[v_indx] += [v_indx]
                    if lmuvlist[v_indx] != []:
                        if lmuvlist[v_indx] != lmuv:
                        #Must be a new LMUV!
                            #duplicate this vert/uv
                            vlist[new_index] = vlist[v_indx]
                            #add its uv
                            uvlist[new_index] = uv
                            lmuvlist[new_index] = lmuv
                            #add this index to a list so we can find it later
                            dupverts[v_indx] += [new_index]
                            v_indx = new_index
                            new_index += 1
                        #else:  Do nothing, we're all good!
                    else:
                            #first time we've seen this.  Add it.
                            lmuvlist[v_indx] = lmuv
                    
                
            else:
                if uvlist[v_indx] != []:
                    rootindx = dupverts[v_indx][0]
                    matches = []
                    for ckvert in dupverts[v_indx]:
                        if uvlist[ckvert] == uv:
                            matches += [ckvert]
                    if len(matches) == 0:
                    #must be a new UV!
                        #duplicate existing vertex
                        vlist[new_index] = vlist[rootindx]
                        #add its new uv
                        uvlist[new_index] = uv
                        #add this index to a list so we can find it later
                        dupverts[rootindx] += [new_index]
                        v_indx = new_index
                        new_index += 1
                    else:
                        v_indx = min(matches)
                else:
                    uvlist[v_indx] = uv
                    dupverts[v_indx] += [v_indx]
                
            index_list += [v_indx]
            
    #re-key to remove gaps from any dropped/invalid verts
    indexlookup = {}
    newindex = 0
    tempindexlist=[]
    vkeys = vlist.keys()
#    print('vkeys=',vkeys)
#    print(dir(vkeys))
#    vkeys.sort()   # Is this necessary ?  Flag this for future.
    for k in vkeys:
        indexlookup[k]=newindex
        newindex += 1
        
    #reindex index_list
    for i in index_list:
        tempindexlist.append(indexlookup[i])
    index_list = tempindexlist[:]
    
    # ********** write face count **************  
    faces = int(len(index_list)/3)
    stream.write(struct.pack("L", faces))
    # ********** write vertex count ************
    stream.write(struct.pack("L", len(vlist)))
    # ************ write mesh flags ************
    stream.write(struct.pack("BBBB", meshflags1, meshflags2, bone_per_vertex, unkflag))
    if has_meshid:
        stream.write(struct.pack("LlL", indexid, vertexid, skeletonid))

    # ************** write ranges *************
    stream.write(struct.pack("f", xr.min))
    stream.write(struct.pack("f", yr.min))
    stream.write(struct.pack("f", zr.min))
    stream.write(struct.pack("f", xr.max))
    stream.write(struct.pack("f", yr.max))
    stream.write(struct.pack("f", zr.max))
    stream.write(struct.pack("f", txr.min))
    stream.write(struct.pack("f", 1 - tyr.min))
    stream.write(struct.pack("f", txr.max))
    stream.write(struct.pack("f", 1 - tyr.max))
    # ************ write lmap range *************
    stream.write(struct.pack("f", lmxr.min))
    stream.write(struct.pack("f", 1 - lmyr.min))
    stream.write(struct.pack("f", lmxr.max))
    stream.write(struct.pack("f", 1 - lmyr.max))
    # *********** write neckdata count **********
    print('Ready to write neckconnect_count')
    print('neckconnect_count = ',neckconnect_count)
    print('has_neckconnect = ',has_neckconnect)
    if has_neckconnect:
        ncpos = stream.tell()
        stream.write(struct.pack("L",neckconnect_count))
#    else:
#        stream.write(struct.pack("L",neckconnect_count))

    # ****** write face list*******
    for tri in index_list:
        stream.write(struct.pack("H", tri))

    if faces & 1:
#        stream.write("\x00\x00")
        stream.write(struct.pack("H", tri))

        
    # ********** write vertices **************
    vkeys = vlist.keys()
#    vkeys.sort()  #  another call to sort.  bypass for now.
    print('Start writing vertices at ',stream.tell())
    for k in vkeys:
        v = vlist[k]
        stream.write(struct.pack("HHHHHH", v[0], v[1], v[2], v[3], v[4], v[5]))
        #write uv
        stream.write(struct.pack("HH", uvlist[k][0], uvlist[k][1]))

    # Lightmap UVs
    if has_lightmap:
        print('has lightmap')
        for k in vkeys:
            stream.write(struct.pack("HH", lmuvlist[k][0], lmuvlist[k][1]))
    print('is_boned = ',is_boned)            
    if is_boned:
        print('start writing weights at ',stream.tell())
#        print('uv_layer = ',mobj.uv_layers[-1])
#        mobj.activeUVLayer = 'UVTex'
        mobj.uv_layers[-1].name = 'UVTex'
        # build vert weight list
        weights = {}
#        print('dupverts = ',dupverts.items())
        for i,dup in dupverts.items():
            for dv in dup:
#                weights[dv] = mobj.getVertexInfluences(i)
                bw_pair = []
                groups = mobj.vertices[i].groups
                vtx = mobj.vertices[i]
                for nbw in range(len(groups)):
                    idx = vtx.groups[nbw].group
                    name = ob.vertex_groups[idx].name
                    bw_pair += [name,groups[nbw].weight]
                weights[dv] = [bw_pair]

                
        # ********** write bone weights **************
        # w0,w1,w2,w3, b0,b1,b2,b3
        ### bone_lookup not initialized!!!!!
        for obj in bpy.data.objects:
            if obj.type == 'ARMATURE':
                arm = obj
#        bone_lookup = arm['']['BoneOrder'].values()
#        print('bone_lookup ',bone_lookup)
        for k in vkeys:
            # fetch the bone/weight pair
            vinf = weights[k]
#            print('vinf = ',vinf)
#            bw_pair = filter(lambda x: x[0] !='neckconnect', vinf)
            bw_pair = vinf[0]
#            print('bw_pair = ',bw_pair)
#            bw_pair.sort(key=lambda x:-x[1])
            bonesort = []
            len_bw_pair = 0
            len_bw_pair = int(len(bw_pair)/2)
#            print('len_bw_pair = ',len_bw_pair)

            for i in range(len_bw_pair):
                bw = [bw_pair[2*i],bw_pair[2*i+1]]
#                print('bw_pair = ',bw[0],bw[1])
                if not(bw[0] == 'neckconnect'):
                    boneref = bone_lookup.index(bw[0])
                    bonesort.append([bw[1], boneref, bw[0]])
 #           bonesort.sort(lambda x,y: cmp(float(y[0]),float(x[0])) or cmp(y[1],x[1]))
#            print('len(bonesort) = ',len(bonesort))
#            print('bonesort = ,',bonesort)
#            print('len_bw_pair = ',len_bw_pair)
            for bw_cnt in range(4):
                if len(bonesort) > bw_cnt:
#                    print('output  = ',bw_cnt,bonesort[bw_cnt][0])
                    stream.write(struct.pack("f", bonesort[bw_cnt][0]))
                else:
#                    print('writing zeros',bw_cnt)
                    stream.write(struct.pack("f", 0))
#            print('len(bonesort) = ',len(bonesort))       
            for bw_cnt in range(4):
#                print('last bone')
                if (bw_cnt < len(bonesort)):
                    if len_bw_pair > bw_cnt:
                        last_bone = bonesort[bw_cnt][1]
#                print('last_bone = ',last_bone)
                stream.write(struct.pack("b", last_bone))  
        print('finsihed writing weights at ',stream.tell())        
    # write Neck Data
    if has_neckconnect:
        print('writing neck data ',neckconnect_count,' at ',stream.tell())
        if neckconnect_count > 0:
            if mshType == "FemaleCostume":
                neckpts = tmConst.NECK_FEMALE
            elif mshType == "MaleCostume":
                neckpts = tmConst.NECK_MALE
            else:
                print("WARNING: mshType is not set to either \'FemaleCostume\' or \'MaleCostume\' -- defaulting to Male neckpoints.")
                neckpts = tmConst.NECK_MALE
            npts = {}
            print('ob = ',ob.name,' type = ',ob.type)
            if('neckconnect' in ob.vertex_groups.keys()):
                print('loading neckconnect points ')
                for n in getVertsFromGroup('neckconnect',ob):
                    print('n = ',n)
                    root = dupverts[n.index][0]
                    print('root = ',root)
                    for pt in dupverts[n.index]:
                        print('neckpts = ',neckpts)
                        print('mobj.vertices[root].co = ',mobj.vertices[root].co)
                        print(getNeckIndex(mobj.vertices[root].co, neckpts))
                        npts[pt] = getNeckIndex(mobj.vertices[root].co, neckpts)
            #backpatch neckpoints count
#            cpos = stream.tell()
#            stream.seek(ncpos)
#            stream.write(struct.pack("L",len(npts)))
#            stream.seek(cpos)
            #write it!
            print('npts.items() = ',npts.items())
            for pt,co in npts.items():
                print('pt = ',pt,' co = ',co)
                stream.write(struct.pack("LL", pt, co))
    print('Completed mesh write for ',ob.name,' at ',stream.tell(),'\n')
def write_armature(stream):
    if has_bones:
#        a = Armature.Get("Bones")
        #bonelist = a.bones.values() -- USE remapped list
#        bone_list = bpy.data.armatures[0].bones
        for obj in bpy.data.objects:
            if obj.type == 'ARMATURE':
                ob = obj
        if 'BoneOrder' in ob:
            boneorder = ob['BoneOrder'].values()
        else:
            boneorder =ob['TheMovies']['BoneOrder'].values()
        if (len(bone_list)):
            #write_null("L", stream)
            #Skeleton ID
            stream.write(struct.pack("L", rigID))
            stream.write(struct.pack("L", len(bone_list)))
              
            #bp_lookup = []
#            for bone in bone_list:
            for name in boneorder:
                bone = ob.data.bones[name]
                #bp_lookup.append(bone.name)
                boneparent = -1
                if bone.parent:
                    boneparent = bone_lookup.index(bone.parent.name)
                write_bone(stream, bone, boneparent)


def write_bone(stream, bone, boneparent):
    write_nts(bone.name, stream)
    stream.write(struct.pack("l", boneparent))

#    bone_matrix = bone.matrix["ARMATURESPACE"]
    bone_matrix = bone.matrix_local
    stream.write(struct.pack("fff", bone_matrix[0][1], bone_matrix[1][1], bone_matrix[2][1]))
    stream.write(struct.pack("fff", bone_matrix[0][2], bone_matrix[1][2], bone_matrix[2][2]))
    stream.write(struct.pack("fff", bone_matrix[0][0], bone_matrix[1][0], bone_matrix[2][0]))
    stream.write(struct.pack("fff", bone_matrix[0][3], bone_matrix[1][3], bone_matrix[2][3]))
            
            
def write_anchor(anchor, stream):
    if 'TheMovies' in anchor:
        if 'ancName' in anchor['TheMovies']:
            ancname = anchor['TheMovies']['ancName']
        elif 'ancName' in anchor:
            ancname = anchor['ancName']
        else:
            ancname = 'anchor'
            print('FAIL: No anchor name property given. using "anchor"')
    else:
        if 'ancName' in anchor:
            ancname = anchor['ancName']
        else:
            ancname = 'anchor'
            print('FAIL: No anchor name property given. using "anchor"')
    write_nts(ancname, stream)
    anc_matrix = anchor.matrix_world
#    anc_matrix.transpose()
    print('anchor = ',ancname)
    m = anc_matrix
    for i in range(4):
        print('%8.6f %8.6f %8.6f %8.6f'%(m[i][0],m[i][1],m[i][2],m[i][3]))
    
    stream.write(struct.pack("fff", anc_matrix[0][0], anc_matrix[1][0], anc_matrix[2][0]))
    stream.write(struct.pack("fff", anc_matrix[0][1], anc_matrix[1][1], anc_matrix[2][1]))
    stream.write(struct.pack("fff", anc_matrix[0][2], anc_matrix[1][2], anc_matrix[2][2]))
    stream.write(struct.pack("fff", anc_matrix[0][3], anc_matrix[1][3], anc_matrix[2][3]))
    
def write_shape(shape, stream):
    stream.write(struct.pack("L",0))
    sh_matrix = shape.matrixWorld
    stream.write(struct.pack("fff", sh_matrix[0][0], sh_matrix[0][1], sh_matrix[0][2]))
    stream.write(struct.pack("fff", sh_matrix[1][0], sh_matrix[1][1], sh_matrix[1][2]))
    stream.write(struct.pack("fff", sh_matrix[2][0], sh_matrix[2][1], sh_matrix[2][2]))
    stream.write(struct.pack("fff", sh_matrix[3][0], sh_matrix[3][1], sh_matrix[3][2]))
    
    bb = calc_bbox(shape.getBoundBox(1))
    
    Sizex = (0 - bb[0]) + bb[1]
    Sizey = (0 - bb[2]) + bb[3]
    Sizez = (0 - bb[4]) + bb[5]
    
    stream.write(struct.pack("fff", Sizex, Sizey, Sizez))
    
#***********************************************
# main export function
#***********************************************

def save_msh(filename):
    '''This is what writes to the msh file, or at least defines the order.
    '''
    
    global material_list, material_keys, Bounds, mshType, bone_list, bone_lookup, rigID, has_bones
    global clickable, min_outline, z_height, lot_boundary, neg_space, shadow, collision,image_list
    
    time1 = time()  #for timing purposes
    prev_time = time1

    #Get Scene ID Properties
#    scn = Blender.Scene.GetCurrent()
    scn = bpy.data.scenes[0]
    Bounds = [0,0,0,0,0,0] 
    mshflags1 = 0
    mshflags2 = 0 
    has_named_groups = 0
    has_static_anim = 0
    has_childmesh = 0
    has_rooms = 0
    has_shapes = 0
    if 'TheMovies' in scn:
        if 'mshName' in scn:
            mshName = scn['mshName']
            mshType = scn['mshType']
            mshflags1 |= ((scn['hide_actor_head'] % 2) * 8)
            mshflags1 |= ((scn['hide_actor_hair'] % 2) * 16)
            mshflags1 |= ((scn['is_auto_animated'] % 2) * 32)
            mshflags1 |= ((scn['tex_replace_mode'] % 2) * 64)
            has_static_anim = scn['has_static_anim'] % 2
            has_shapes = scn['has_shapes'] % 2
            has_childmesh = scn['has_childmesh'] % 2
            has_rooms = scn['has_rooms'] % 2
        else:
            mshName = scn['TheMovies']['mshName']
            mshType = scn['TheMovies']['mshType']
            mshflags1 |= ((scn['TheMovies']['hide_actor_head'] % 2) * 8)
            mshflags1 |= ((scn['TheMovies']['hide_actor_hair'] % 2) * 16)
            mshflags1 |= ((scn['TheMovies']['is_auto_animated'] % 2) * 32)
            mshflags1 |= ((scn['TheMovies']['tex_replace_mode'] % 2) * 64)
            has_static_anim = scn['TheMovies']['has_static_anim'] % 2
            has_shapes = scn['TheMovies']['has_shapes'] % 2
            has_childmesh = scn['TheMovies']['has_childmesh'] % 2
            has_rooms = scn['TheMovies']['has_rooms'] % 2            
    else: 
        mshName = "p_export.msh"
        mshType = "Undefined"
        
    image_list = []
#    for i in Image.Get():
#    for i in bpy.data.images:
#       image_list.append(i.name) # probably need to eliminate textures not referenced in Materials...
    obkeys = bpy.data.objects.keys()
    obkeys.sort()
    for obkey in obkeys:
#    for ob in bpy.data.objects:
        ob = bpy.data.objects[obkey]
        if ob.type == 'MESH':
            for mat in ob.data.materials:
                if mat != None:
                    for i in range(4):
                        ts = mat.texture_slots[i]
                        if mat.texture_slots[i] == None:
                            print('Empty slot')
                        else:
                            if mat.use_textures[i]:
                                if (ts.texture == None ) == False:
                                    if ts.texture.image != None:
                                        image = ts.texture.image.name
                                        if image not in image_list:
                                            image_list.append(ts.texture.image.name)
    clickable = None
    min_outline = None
    z_height = None
    lot_boundary = None
    neg_space = None
    shadow = None
    collision = None
    
    group_names = []
    group_values = []
    group_orderkeys=[]
    anchor_list = []
    shape_list = []
    room_list = {}
    bone_list = []
    bone_lookup = []
    
    # Groups
    rigID = 0
##    bbox_all=[]
    #sort group order keys
    for g in bpy.data.groups:
        group_orderkeys.append(g.name)
#    group_orderkeys.sort()
    for g in group_orderkeys:
        gr = bpy.data.groups[g]
        if gr.name == 'control_meshes':
            for m in gr.objects:
                if m.getType() == "Mesh":
                    n = m.name
                    Bounds = adjustBounds(m)
                    if n == "clickable":
                        clickable = m
                    elif n == "collision":
                        mshflags2 |= 16
                        collision = m
                    elif n == "shadow":
                        mshflags2 |= 8
                        shadow = m
                    elif n == "z_height":
                        mshflags1 |= 4
                        z_height = m
                    elif n == "neg_space":
                        mshflags2 |= 4
                        neg_space = m
                    elif n == "min_outline":
                        mshflags1 |= 2
                        min_outline = m
                    elif n == "lot_boundary":
                        mshflags2 |= 1
                        lot_boundary = m
        elif gr.name == 'rooms':
            for m in gr.objects:
                if m.getType() == "Mesh":
                    if m.properties.has_key('TheMovies'):
                        n = m.properties['roomName']
                    else:
                        n = m.name
                    has_rooms = 1
                    Bounds = adjustBounds(m)
                    room_list[n] = m
        elif gr.name == 'shapes':
            for m in gr.objects:
                if m.getType() == "Mesh":
                    Bounds = adjustBounds(m)
                    has_shapes = 1
                    shape_list.append(m)
        elif gr.name == 'anchors':
            for o in gr.objects:
                if o.type == 'EMPTY':
                    if o.name != "Anchors":
                        anchor_list.append(o)
                        Bounds = adjustBounds(o)
        else:
            emptyfound = False
            print('gr.name = ',gr.name)
            print('gr.objects: ',gr.objects.values())
            print('gr.objects.keys = ',gr.objects.keys())
            for o in gr.objects:
                print('o.name = ',o.name)
                if o.type == "EMPTY":
                    emptyfound = True
                    try:
                        grpname = o['grpName']
                    except:
                        #great.  Why are you trying to break this??  We'll try to adjust.
                        print("WARNING: No \'grpName\' ID Property found in Empty \'%s\'.  Using Empty name." % (o.name))
                        grpname = o.name.split('.',1)[-1]
                    Bounds = adjustBounds(o)
                if o.type == "MESH":
                    print('MESH Name = ',o.name)
                    mats = o.material_slots
                    if mats:
                        mat = mats[0]
                        print('mat_name = ',mat.name)
                        if mat.name not in material_keys:
                            material_list.append(mat)
                            material_keys.append(mat.name)
                    Bounds = adjustBounds(o)
            if emptyfound == False:
                grpname = gr.name
                print("FAIL! (Blender Group \'%s\') Contains no Empty pivot!" % grpname)
            if has_named_groups == 0:
                r = unnamed_group.match(grpname)
                if not r:
                    print('named group = ',grpname)
                    has_named_groups = 1
            if has_static_anim == 0 :
                r2 = static_anim_group.match(grpname)
                if r2:
                    has_static_anim = 1
                    
            #group_list[grpname] = gr.objects
            group_names.append(grpname)
            group_values.append(gr.objects)
##    bbox = calc_bbox(bbox_all)
    

#    if Blender.Armature.Get() != {} :
#    if 'Armature' in bpy.data.armatures.keys():
    if len(bpy.data.armatures.keys()) > 0:
        has_bones = 1
        bone_list = []
        
        for ob in scn.objects:
            if ob.type == 'ARMATURE':
                arm = ob
                try:
                    rigID = int(ob.name)
                except:
                    try:
                        skt = ob.properties['skeletonType']
                        for key in tmConst.RIG_IDS.keys():
                            value = tmConRIG_IDS[key]
                            if not index.has_key(value):
                                index[value] = key
                        try:
                            rigID = index[skt]
                            del(index)
                        except:
                            print("WARNING: skeletonType not recognized. Generating random Skeleton ID")
                            rigID = randint(0,4294967295)
                    except:
                        print("WARNING: skeletonType property missing. Generating random Skeleton ID")
                        rigID = randint(0,4294967295)
             
        if 'TheMovies' in arm:
            bone_lookup = arm['TheMovies']['BoneOrder'].values()
            for bname in bone_lookup:
                b = bpy.data.armatures[0].bones[bname]
                bone_list.append(b)
        elif 'BoneOrder' in arm:
            bone_lookup = arm['BoneOrder'].values()
            for bname in bone_lookup:
                b = bpy.data.armatures[0].bones[bname]
                bone_list.append(b)
        else:
            # Fallback .. Mesh hasn't got ordered Armature.  We'll give them a fighting chance...
            print("WARNING: Bone Order not found.  Proceeding to output in Undefined order.  Mesh could have issues.")
            rootbone = Armature.Get("Bones").bones["root"]
            bonekeys = rootbone.getAllChildren()
            bonekeys.insert(0, rootbone)
            for bone in bonekeys:
                bone_list.append(bone)
            
            # Tarison discovered these rigs need the Headbone to be at the end of the bone list
            # we may need to add other exceptions, maybe for all 'costume' classed meshes too
            # remove headbone (and any children)
            if rigID == 507812352 or rigID == 1094395392 : 
                headstartbone = Armature.Get("Bones").bones["neck"] #get all bones attached to neck.. i.e. head and all childbones
                headchildren = headstartbone.getAllChildren()           
                for hc in headchildren:
                    bone_list.remove(hc)
                for hc in headchildren:
                    bone_list.append(hc)
            
            for bone in bone_list:
                bone_lookup.append(bone.name)
        
    else:
        has_bones = 0
       
    mshflags2 |= (has_shapes * 2)
    mshflags2 |= (has_childmesh * 64)
    mshflags2 |= (has_rooms * 128)
    childmesh_list = [] #never used?

    #*******************************************
    # write header
    #*******************************************
    stream = open(filename, 'wb')
    print('write header at ',stream.tell())
    stream.write(struct.pack("L",10))
    
    # image count 
    stream.write(struct.pack("L", len(image_list)))
    # material count 
    stream.write(struct.pack("L", len(material_keys)))
    # group count 
    stream.write(struct.pack("L",len(group_values)))
    # has bones
    stream.write(struct.pack("B", has_bones))
    # static anims
    stream.write(struct.pack("B", has_static_anim))
    # anchor count
    stream.write(struct.pack("B", len(anchor_list)))
    # clickable mesh
    if clickable is not None:
        stream.write(bytes([1]))
    else:
        stream.write(bytes([0]))
    #  has outline
    stream.write(bytes([0]))
    # flags 1
    stream.write(struct.pack("B", mshflags1))
    # flags 2
    stream.write(struct.pack("B", mshflags2))
    # named groups
    stream.write(struct.pack("B", has_named_groups))
    # shape count
    print('shape count = ',len(shape_list))
    print('at ',stream.tell())
    if has_shapes:
        stream.write(struct.pack("L", len(shape_list)))
    # childmesh (?)
    if has_childmesh:
        stream.write(struct.pack("L", len(childmesh_list))) 
    # room count
    if has_rooms:
        stream.write(struct.pack("L", len(room_list)))
    # end header **************************
    print('header written : %.4f sec.' % (time()-prev_time))
    prev_time = time()
    
    # image list
    
    print('writing image_list')
    image_list = bpy.data.images
    for i in range(len(image_list)):
        fp = bpy.data.images[i].filepath
        print('fp = ',fp)
#        im = os.path.split(fp)[1]
        im = image_list[i].name
        print('im = ',im)
        write_nts(im, stream)
    print('%d images written : %.4f sec.' % (len(image_list), time()-prev_time))
    prev_time = time()
    
        
    # material list
    print('material_list = ',material_list)
    print('Number of materials = ',len(material_list))
#    material_list = bpy.data.materials
    print('writing materials at ',stream.tell())
    for mat in material_list:
        print('mat = ',mat.name)
        write_material(mat.material, stream)
    print('%d materials written : %.4f sec.' % (len(material_list), time()-prev_time))
    prev_time = time()

    # control meshes
    print('write control meshes at ',stream.tell())
    write_control_meshes(stream)
    print('control_meshes written : %.4f sec.' % (time()-prev_time))
    prev_time = time()

    # rooms
    for n in room_list:
        write_nts(n, stream)
        write_basic_mesh(room_list[n], stream)
    print('%d rooms written : %.4f sec.' % (len(room_list), time()-prev_time))
    prev_time = time()

    # groups - mesh contents
    print(' writing groups')
    print('group_values = ',group_values)
#    for g in reversed(group_values):
    for g in group_values:
        write_group(g, stream)
        print('group written : %.4f sec.' % (time()-prev_time))
        prev_time = time()
        

    # bones goes here
    write_armature(stream)
    print('armatures written : %.4f sec.' % (time()-prev_time))
    prev_time = time()

    # group names
    for g in group_names:
        write_nts(g, stream)
    print('%d group_names written : %.4f sec.' % (len(group_names), time()-prev_time))
    prev_time = time()
    
    # Anchors
    for a in anchor_list:
        write_anchor(a, stream)
    print('%d anchors written : %.4f sec.' % (len(anchor_list), time()-prev_time))
    prev_time = time()
    
    # Convex Hull (We're skipping this piece of crap section for now!)
    
    # Shapes
    # not sure what they're for, but we can write them!
    for s in shape_list:
        write_shape(s, stream)
    print('%d shapes written : %.4f sec.' % (len(shape_list), time()-prev_time))
    prev_time = time()
        
    stream.close()
    print('has_bones = ',has_bones)
    print('finished exporting: "%s" in %.4f sec.' % (filename, (time()-time1)))
#***********************************************
# register callback
#***********************************************
#def my_callback(filename):
#    save_msh(filename)

#Blender.Window.FileSelector(my_callback, "Export MSH", '*.msh')
class TheMoviesMeshExporter(Operator, ExportHelper):
    """Export a TM msh object"""
    bl_idname = "export_test.tm_msh"
    bl_label = "Export TM MSH"
    bl_options = {'REGISTER','UNDO'}
    filename_ext = ".msh"
    filter_glob = StringProperty(
        default="*.msh",
        options= {'HIDDEN'},
        maxlen=255,
        )
    use_setting = BoolProperty(
        name="Example Boolean",
        description="Example Tooltip",
        default=True,
        )
    type = EnumProperty(
        name="Example Enum",
        description="Choose between two items",
        items=(('OPT_A',"First Option","Description one"),
               ('OPT_B',"Second Option","Description two")),
        default='OPT_A',
        )
    def execute(self,context):
        save_msh(self.filepath)
        return {'FINISHED'}
def menu_func_export(self,context):
        self.layout.operator(TheMoviesMeshExporter.bl_idname, text="Export TM Mesh (*.msh)")
#        bpy.types.INFO_MT_file_export.remove(menu_func_export)
def register():
        bpy.utils.register_class(TheMoviesMeshExporter)
        bpy.types.INFO_MT_file_export.append(menu_func_export)
def mshout():
    bpy.ops.export_test.tm_msh('INVOKE_DEFAULT')

if __name__ == "__main__":
    register()
#    bpy.ops.export_test.tm_msh('INVOKE_DEFAULT')
    
