# ***** BEGIN GPL LICENSE BLOCK *****
#
# Script copyright (C) Mark S Andrews, Nick Hudson, Glen Rickey, Cary Semar
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****
# --------------------------------------------------------------------------
#  Script name:
#  msh_import_1.8.py
#  Capabilities:
#       This script will import a costume msh or prop msh from "The Movies."
#  Limitations:
#       Render view works, texture view does not work
#       set_*.msh will not work.
#
#   Converted to Blender 2.79b by Cary Semar 1/1/2022
#                                  Released: 1/3/2022
#
#   2/18/2022:  set parent and keep transform
#   mesh.select = True
#   gr.select = True
#   bpy.context.scene.objects.active = gr
#   bpy.ops.object.parent_set(type='OBJECT',xmirror=False,keep_transform=True)
#   Deselect objects after setting parent.
#
bl_info = {"name": "The Movies Mesh Import", 
    "description": "Imports models from The Movies",
    "category": "Import-Export",
    "blender": (2,79,2),
    "location": "File > Import > The Movies Mesh (.msh)",
    "author": "Nick Hudson, Glen Rickey, Mark S Andrews, Cary Semar"
}

import bpy, os, struct, math, bmesh
from time import time
from bpy_extras.io_utils import ImportHelper
from bpy_extras.object_utils import object_data_add
from bpy.props import StringProperty, BoolProperty, EnumProperty
from bpy.types import Operator
import mathutils
import tmConst
from mathutils import Vector, Matrix
r2d = 180.0/math.pi
print('Installed: tm_msh_import ver 3.3')
#import pdb
def vec_roll_to_mat3(vec, roll):
    target = mathutils.Vector((0,1,0))
    nor = vec.normalized()
    axis = target.cross(nor)
    if axis.dot(axis) > 0.0000000001: # this seems to be the problem for some bones, no idea how to fix
        axis.normalize()
        theta = target.angle(nor)
        bMatrix = mathutils.Matrix.Rotation(theta, 3, axis)
    else:
        updown = 1 if target.dot(nor) > 0 else -1
        bMatrix = mathutils.Matrix.Scale(updown, 3)
    rMatrix = mathutils.Matrix.Rotation(roll, 3, nor)
    mat = rMatrix * bMatrix
    return mat
def mat3_to_vec_roll(mat):
    vec = mat.col[1]
    vecmat = vec_roll_to_mat3(mat.col[1], 0)
    vecmatinv = vecmat.inverted()
    rollmat = vecmatinv * mat
    roll = math.atan2(rollmat[0][2], rollmat[2][2])
    return vec, roll

#***********************************************
# globals
#***********************************************
image_list = []
material_list=[]
neckdata = []
lmuv_list = []
"""
try:
    group_controlmesh = bpy.data.groups.get('control_meshes') 
except:
    group_controlmesh = bpy.data.groups.new('control_meshes')
try:
    group_room = bpy.data.groups.get('rooms') 
except:
    group_room = bpy.data.groups.new('rooms')
try:
    group_shapes = bpy.data.groups.get('shapes')
except:
    group_shapes = bpy.data.groups.new('shapes')

try:
    group_anchors = bpy.data.groups.get('anchors')
except: 
    group_anchors = bpy.data.groups.new('anchors')
group_anchors = bpy.data.groups.get('anchors')
if group_anchors == None:
    group_anchors = bpy.data.groups.new('anchors')
"""
#***********************************************
# helper classes & functions for msh file
#***********************************************
def get_texture(name):
    textures = bpy.data.textures
    for tex in textures:
        if tex.name == name:
            return tex
    return None
def get_material(name):
    materials = bpy.data.materials
    for mat in materials:
        if mat.name == name:
            return mat
    return None
def read_item(struct_type, stream):
    value = struct.unpack(struct_type,
        stream.read(struct.calcsize(struct_type)))
    if len(value) == 1:
        value = value[0]
    return value
def read_nts(stream, length=32):
    s = stream.read(length)
#    print('read_nts = ',stream.tell())
    return s.decode("ASCII")

class decoder(object):
    _struct = ''
    _fields = {}
    _flags = {}
    
    def __init__(self, stream):
        self._value = struct.unpack(self._struct,
            stream.read(struct.calcsize(self._struct)))
    
    def __getattr__(self, item):
        if item in self._fields.keys():
            keyid = self._fields[item]
            return self._value[keyid]
        elif item in self._flags.keys():
            flag, bit = self._flags[item]
            return bool(self._value[flag] & bit)
        else:
            raise AttributeError(item)

    def __repr__(self):
        return repr(self._value)

class range_adjust(object):
    def __init__(self, min, max):
        self.min = min
        self.max = max
        self.step = (max - min) / 65535.0
        object.__init__(self)
    def adjust(self, value):
        return self.min + (value * self.step)

def get_image_path(basepath, filename):
#    pdb.set_trace()
    p = os.path.join(basepath, filename)
    if os.path.isfile(p):
        return p
    else:
        dirs = ["lightmap","props","costumes","accessories",
            "hair","makeup","people","backdrops"]
        for dir in dirs:
            p = os.path.join(basepath, dir, filename)
            if os.path.isfile(p):
                return p

#***********************************************
# class definitions for msh file
#***********************************************

class anchor(decoder):
    _struct = "ffffffffffff"
    _fields = {"m11":0,
        "m12":1,
        "m13":2,
        "m21":3,
        "m22":4,
        "m23":5,
        "m31":6,
        "m32":7,
        "m33":8,
        "offset_x":9,
        "offset_y":10,
        "offset_z":11,
        }
    
    def __init__(self, stream):
        self.anchor_name = read_nts(stream)
        decoder.__init__(self, stream)
        m1 = [self.m11, self.m12, self.m13]
        m2 = [self.m21, self.m22, self.m23]
        m3 = [self.m31, self.m32, self.m33]
        m4 = [self.offset_x, self.offset_y, self.offset_z]
        self.matrix = Matrix((m1, m2, m3, m4)).resize_4x4()

class bone(decoder):
    _struct = "lffffffffffff"
    _fields = {"parent":0,
        "m11":1,
        "m12":2,
        "m13":3,
        "m21":4,
        "m22":5,
        "m23":6,
        "m31":7,
        "m32":8,
        "m33":9,
        "offset_x":10,
        "offset_y":11,
        "offset_z":12,
        }
    def __init__(self, stream):
        self.bone_name = read_nts(stream)
        decoder.__init__(self, stream)
        m1 = [self.m31, self.m32, self.m33,0.]
        m2 = [self.m11, self.m12, self.m13,0.]
        m3 = [self.m21, self.m22, self.m23,0.]
##        m1 = [self.m11, self.m21, self.m31]
##        m2 = [self.m12, self.m22, self.m32]
##        m3 = [self.m13, self.m23, self.m33]
        m4 = [self.offset_x, self.offset_y, self.offset_z,1.]
        self.matrix = Matrix((m1, m2, m3, m4)) #
        self.head = Vector((m4[0],m4[1],m4[2]))
        
        

class skeleton_header(decoder):
    _struct = "LL"
    _fields = {"rig_id":0,
               "bone_count":1,}

class bone_weights(decoder):
    _struct = "ffffbbbb"
    _fields = {"w0":0,
        "w1":1,
        "w2":2,
        "w3":3,
        "b0":4,
        "b1":5,
        "b2":6,
        "b3":7,
        }

class material(decoder):
    _struct = "bbbbBBBBBBBBBbBxBBBBbbbx"
    _fields = {"map0":0,
        "map1":1,
        "map2":2,
        "map3":3,
        "alphaenvmap":13,
        "color_B":15,
        "color_G":16,
        "color_R":17,
        "color_A":18,
        "scroll_U":19,
        "scroll_V":20,
        "rot_UV":21,
        }
    _flags = {"doublesided":(4,1),
              "floor_shadow_tex":(4,2),
              "alpha_separate":(4,4),
              
              "wrap_U":(5,255),
              "wrap_V":(6,255),
              "use_alpha":(7,255),
              "enable_alpha_test":(8,255),
              "glass":(9,255),
              "water":(10,255),
              "still_water":(11,255),
              "alpha_per_vertex":(12,255),
              
              "invisible":(14,1),
              "not_z_write":(14,2),
              "has_uv":(14,4),
              "self_lit":(14,8),
              "no_floor_shadow":(14,16),
              "no_delaydraw_transp":(14,32),
              "tri_sort":(14,64),
              "additive":(14,128),
              
        }

class compressed_vertex(decoder):
    _struct = "HHHHHHHH"
    _fields = {"x":0,
            "y":1,
            "z":2,
            "nx":3,
            "ny":4,
            "nz":5,
            "tx":6,
            "ty":7,
            }

class msh_header(decoder):
    _struct = "LLLLBBBBxBBB"
    _fields = {"image_count":1,
               "material_count":2,
               "group_count":3,
        
               "anchor_count":6,
        
        }
    _flags = {"has_bones":(4,255),
                "has_static_anim":(5,255),
              
                "has_clickable":(7,255),
              
                "Unk":(8,1),
                "has_min_outline":(8,2),
                "has_z_height":(8,4),
                "hide_actor_head":(8,8),
                "hide_actor_hair":(8,16),
                "is_auto_animated":(8,32),
                "tex_replace_mode":(8,64),
                "has_convex_hull":(8,128),
                
                "has_lot_boundary":(9,1),
                "has_shapes":(9,2),
                "has_neg_space":(9,4),
                "has_shadow":(9,8),
                "has_collision":(9,16),
                "has_blueprint":(9,32),
                "has_childmesh":(9,64),
                "has_rooms":(9,128),
                  
                "has_named_groups":(10,255),
        }
    
    def __init__(self, stream):
        decoder.__init__(self, stream)
        if self._value[0] != 10:
            print("FAIL! Not a valid mesh file")
            return
            
        if self.has_shapes:
            self.shape_count = read_item("L", stream)
        else:
            self.shape_count = 0
            
        if self.has_childmesh:
            self.childmesh_count = read_item("L", stream)
        else:
            self.childmesh_count = 0
            
        if self.has_rooms:
            self.room_count = read_item("L", stream)
        else:
            self.room_count = 0


class mesh_header(decoder):
    _struct = "LLLBBBB"
    _fields = {"materialid":0,
               "face_count":1,
               "vertex_count":2,
               "bone_per_vertex":5,
        }
    _flags = {"has_weights":(3,1),
              "has_floor_reflections":(3,2),
              "no_outline":(3,4),
              "is_landscape":(3,8),
              "has_meshid":(3,16),
              "is_compressed":(3,32),
              "has_neckconnect":(3,64),
              "accepts_actor_shadow":(3,128),
              
              "has_lightmap":(4,1),
              "is_minutehand":(4,2),
              "is_hourhand":(4,4),
              "static_backdrop":(4,8),
              
              "unk_flag":(6,255),
        }
    
    def __init__(self, stream):
        decoder.__init__(self, stream)
        # print "Mesh:", self._value
        if self.has_meshid:
            self.indexid = read_item("l", stream)
            self.vertexid = read_item("l", stream)
            self.skeletonid = read_item("l", stream)
        else:
            self.indexid = None
            self.vertexid = None
            self.skeletonid = None
            
        if self.is_compressed:                
            min_x = read_item("f", stream)
            min_y = read_item("f", stream)
            min_z = read_item("f", stream)
            max_x = read_item("f", stream)
            max_y = read_item("f", stream)
            max_z = read_item("f", stream)
            min_tx = read_item("f", stream)
            min_ty = read_item("f", stream)
            max_tx = read_item("f", stream)
            max_ty = read_item("f", stream)
            min_lm_x = read_item("f", stream)
            min_lm_y = read_item("f", stream)
            max_lm_x = read_item("f", stream)
            max_lm_y = read_item("f", stream)
            x_range = range_adjust(min_x, max_x)
            y_range = range_adjust(min_y, max_y)
            z_range = range_adjust(min_z, max_z)
            tx_range = range_adjust(min_tx, max_tx)
            ty_range = range_adjust(1.0 - min_ty, 1.0 - max_ty)
            lm_x_range = range_adjust(min_lm_x, max_lm_x)
            lm_y_range = range_adjust(1.0 - min_lm_y, 1.0 - max_lm_y) 
            n_range = range_adjust(-1.0, 1.0)
            self.adjust_n = n_range.adjust
            self.adjust_x = x_range.adjust
            self.adjust_y = y_range.adjust
            self.adjust_z = z_range.adjust
            self.adjust_tx = tx_range.adjust
            self.adjust_ty = ty_range.adjust
            self.adjust_lm_x = lm_x_range.adjust
            self.adjust_lm_y = lm_y_range.adjust
        
        if self.has_neckconnect:    
            self.neckconnect_count = read_item("L", stream)
        else:
            self.neckconnect_count = 0

class group_header(decoder):
    _struct = "BBBB"
    _fields = {"mesh_count":0,
        }
    _flags = {"has_transanim":(2,1),
              "is_a_carbody":(2,2),
            
              "is_land":(3,4),
              "hide_reflection":(3,8),
              "has_hidden":(3,16),
              }
    
    def __init__(self, stream):
        decoder.__init__(self, stream)
        #print "Group:", self._value
        matrix = read_item("ffffffffffff", stream) # This had columns.rows reversed... I think that was wrong
        m1 = [matrix[0], matrix[1], matrix[2],0]
        m2 = [matrix[3], matrix[4], matrix[5],0]
        m3 = [matrix[6], matrix[7], matrix[8],0]
        m4 = [matrix[9], matrix[10], matrix[11],1]
        pivot = Matrix((m1, m2, m3, m4))
        self.pivot = pivot
        if self.has_hidden:
            self.hidden_on = read_item("H", stream)
            self.hidden_off = read_item("H", stream)
        else:
            self.hidden_on = 0
            self.hidden_off = 0
        
class shape(decoder):
    _struct = "L"
    _fields = {"unknown":0}
    
    def __init__(self,stream):
        decoder.__init__(self, stream)
        matrix = read_item("ffffffffffff", stream)
        m1 = [matrix[0], matrix[1], matrix[2]]
        m2 = [matrix[3], matrix[4], matrix[5]]
        m3 = [matrix[6], matrix[7], matrix[8]]
        m4 = [matrix[9], matrix[10], matrix[11]]
        self.center = Matrix((m1, m2, m3, m4)).resize_4x4()
        
        self.dimensions = read_item("fff",stream)
            
#***********************************************
# reader functions for msh file
#***********************************************

def add_control_mesh(mesh_name, stream, layer=2):
    try:
        controlmesh = bpy.data.objects.get('ControlMeshes')
    except:
        controlmesh = bpy.data.objects.new('Empty','ControlMeshes')
        group_controlmesh.objects.link(controlmesh)
    vertex_count = read_item("L", stream)
    face_count = read_item("L", stream)
    mesh = bpy.data.meshes.new("control")
    bm = bmesh.new()
    vertex_index = []
    for vertexid in range(vertex_count):
#        p = read_item("fff", stream)
        (x,y,z) = read_item("fff",stream)
#        v = NMesh.Vert(p[0],p[1],p[2])
        v = bm.verts.new((x, y, z))
#        mesh.verts.append(v)
        bm.verts.ensure_lookup_table()
        vertex_index.append(bm.verts[-1])
        vertex_id = len(bm.verts) - 1
    for faceid in range(face_count):
        face_keys = read_item("HHH", stream)
        v0 = vertex_index[face_keys[0]]
        v1 = vertex_index[face_keys[1]]
        v2 = vertex_index[face_keys[2]]
#        f= NMesh.Face([v0, v1, v2])
#        bm.faces.new([v0,v1,v2])
#        mesh.faces.append(f)
    if face_count & 1:
        pad = stream.read(2)
    bm.to_mesh(mesh)
    mesh.update()    
#    ob = NMesh.PutRaw(mesh)
#    controlmesh.makeParent([ob],0,1)
#    ob.name = mesh_name
#    ob.layers = [layer]
#    if mesh_name.startswith("room_"):
#        group_room.objects.link(ob)
#        if ob.properties.has_key('TheMovies')== False:
#            ob.properties['TheMovies'] = {}
#        ob.properties['TheMovies']['roomName'] = mesh_name #store full name because Blender clips them...
#    else:
#        group_controlmesh.objects.link(ob)
#    ob.dloc = [0,0,0]
#    Window.RedrawAll()

def get_mesh(mesh_name, header, stream):
    global material_list, lmuv_list

    face_list = []
    for faceid in range(header.face_count):
        face_list.append(list(read_item("HHH", stream)))
    if header.face_count & 1:
        pad = stream.read(2)

    vertex_list = []
    for vertexid in range(header.vertex_count):
        vertex_list.append(compressed_vertex(stream))
    
    lmuv_list=[]
    if header.has_lightmap:
        for vertexid in range(header.vertex_count):
            lmuv_list.append(list(read_item("HH", stream)))
    
    if header.has_weights:
        for vertexid in range(header.vertex_count):
            vertex_list[vertexid].weights = bone_weights(stream)
  
    
    neckdata = []
    for nid in range(header.neckconnect_count):
        #neckconnect.append(vertex_list[read_item("L",stream)])
        neckdata.append(read_item("L",stream))
        ncindex = read_item("L", stream) #we don't need this, we'll generate new on export
        #we'll generate a vertex group for this after verts are added ;)

    mesh = bpy.data.meshes.new(name=mesh_name)
    bm = bmesh.new()
    
    #mesh.hasFaceUV(True)
    mesh.materials.append(material_list[header.materialid])
    mesh.use_auto_smooth = True
    mesh.show_double_sided = True

    #mesh.setMode("TwoSided", "AutoSmooth")  ## change from 2.49b
    
    # add mesh vertices
    vertex_index = []
    bones = {}
    for cv in vertex_list:
        x = header.adjust_x(cv.x)
        y = header.adjust_y(cv.y)
        z = header.adjust_z(cv.z)
        nx = header.adjust_n(cv.nx)
        ny = header.adjust_n(cv.ny)
        nz = header.adjust_n(cv.nz)
        
        # v = #NMesh.Vert(x, y, z)
        v = bm.verts.new((x, y, z))
        # v = bpy.types.MeshVertex(x, y, z)
        # norm = Vector((nx, ny, nz))
        # v.no[0] = nx
        # v.no[1] = ny
        # v.no[2] = nz
        # v.normal = norm
        # bm.verts.append(v)
        vertex_id = len(bm.verts) - 1

        # This sets the bones/weights
        if "weights" in dir(cv):
            if cv.weights.b0 != -1:
#                print('bones = ',bones)
                if cv.weights.b0 in bones:
#                    if bones[cv.weights.b0].get(cv.weights.w0) != None:
                    if cv.weights.w0 in bones[cv.weights.b0]:
                        bones[cv.weights.b0][cv.weights.w0].append(vertex_id)
                    else:
                        bones[cv.weights.b0][cv.weights.w0]= [vertex_id]
                else:
                        bones[cv.weights.b0] = {cv.weights.w0:[vertex_id],}
            if cv.weights.b1 != -1:
#                if bones.get(cv.weights.b1) != None:
                if cv.weights.b1 in bones:
#                    if bones[cv.weights.b1].get(cv.weights.w1) != None:
                    if cv.weights.w1 in bones[cv.weights.b1]:
                        bones[cv.weights.b1][cv.weights.w1].append(vertex_id)
                    else:
                        bones[cv.weights.b1][cv.weights.w1]= [vertex_id]
                else:
                    bones[cv.weights.b1] = {cv.weights.w1:[vertex_id],}
            if cv.weights.b2 != -1:
                if cv.weights.b2 in bones:
                    if cv.weights.w2 in bones[cv.weights.b2]:
                        bones[cv.weights.b2][cv.weights.w2].append(vertex_id)
                    else:
                        bones[cv.weights.b2][cv.weights.w2]= [vertex_id]
                else:
                    bones[cv.weights.b2] = {cv.weights.w2:[vertex_id],}
            if cv.weights.b3 != -1:
#                if bones.get(cv.weights.b3) != None:
                if cv.weights.b3 in bones:
#                    if bones[cv.weights.b3].get(cv.weights.w3) != None:
                    if cv.weights.w3 in bones[cv.weights.b3]:
                        bones[cv.weights.b3][cv.weights.w3].append(vertex_id)
                    else:
                        bones[cv.weights.b3][cv.weights.w3]= [vertex_id]
                else:
                    bones[cv.weights.b3] = {cv.weights.w3:[vertex_id],}

        bm.verts.ensure_lookup_table()
        vertex_index.append(bm.verts[-1])
                        
    # add textured faces
    fuvw = {}
    fn = 0
    for face in face_list:
        v0 = vertex_index[face[0]]
        tx0 = header.adjust_tx(vertex_list[face[0]].tx)
        ty0 = header.adjust_ty(vertex_list[face[0]].ty)
        v1 = vertex_index[face[1]]
        tx1 = header.adjust_tx(vertex_list[face[1]].tx)
        ty1 = header.adjust_ty(vertex_list[face[1]].ty)
        v2 = vertex_index[face[2]]
        tx2 = header.adjust_tx(vertex_list[face[2]].tx)
        ty2 = header.adjust_ty(vertex_list[face[2]].ty)
        fuvw[fn] = [(tx0,ty0),(tx1,ty1),(tx2,ty2)]
        fn = fn + 1
        try:
            bm.faces.new([v0,v1,v2])
        except:
            print('face already exists:',(v0,v1,v2))
    uv_layer = bm.loops.layers.uv.verify()
    fn = 0
    for f in bm.faces:
        ln = 0
        uvwdata = fuvw[fn]
        for l in f.loops:
            luv = l[uv_layer]
            (x,y) = (luv.uv[0],luv.uv[1])
            (tx,ty) = uvwdata[ln]
            luv.uv = Vector((tx,ty))
            ln = ln + 1
        fn = fn + 1           
    bm.to_mesh(mesh)
    mesh.update()
    
#    for f in mesh.faces:
#        f.uv = [(tx0, ty0), (tx1,ty1), (tx2,ty2)]
#        im = bpy.data.materials.get()[header.materialid].getTextures()[0].tex.getImage()
#        if im != None:
#            f.image = im
        
        
    #mesh.foreach_set("use_smooth", [True] * len(mesh.polygons))
    
    #todo - smoothing angle...
    #for face in mesh.faces:
    #    face.smooth = 1
        
    return mesh, bones, neckdata

#***********************************************
# load msh file
#***********************************************

def load_msh (filename):
    global image_list, material_list, group_controlmesh, lmuv_list
    start_time = time()
    prev_time = start_time
    image_list = []
    material_list=[]
    neckdata = []
    lmuv_list = []
    
#    bpy.ops.object.mode_set(0)
# check poll() to avoid exception
    if bpy.ops.object.mode_set.poll():
        bpy.ops.object.mode_set(mode='OBJECT')
#    in_editmode = Window.EditMode()
#    if in_editmode:
#        Window.EditMode(0)
    stream = open(filename,"rb")
    filepath = os.path.dirname(filename)
    filebase = os.path.basename(filename)
    fullbase = os.path.splitext(filebase)[0]
    print('filepath=',filepath)
    print('filebase=',filebase)
    print('fullbase=',fullbase)
    basename = fullbase[0:21] #Blender has a $%$@ 21-char limit!
    basedir = os.path.split(filepath)[0]
    imagedir = os.path.join(basedir,"textures")
    print('basename=',basename)
    print('imagedir=',imagedir)
    
    scene = bpy.data.scenes[0]
    # scene.setLayers([1])
    header = msh_header(stream)
    
    ##  populate Scene ID Properties    
    mshType = 'Undefined'
    
    if ('TheMovies' in scene) == False:
        scene['TheMovies'] = 'Scene'

            
    scene['mshName'] = fullbase
    if fullbase.startswith("set_") or fullbase.startswith("sld_"):
        mshType = 'Set'
    if fullbase.startswith("fac_") or fullbase.startswith("fld_"):
        mshType = 'Facility'
    if fullbase.startswith("p_car_") or fullbase.startswith("lp_car_"):
        mshType = 'Car'
    if fullbase.startswith("p_") or fullbase.startswith("lp_"):
        mshType = 'Prop'
    if fullbase.startswith("cos_f_") :
        mshType = 'FemaleCostume'
    if fullbase.startswith("cos_m_") :
        mshType = 'MaleCostume'
    if fullbase.startswith("acc_") :
        mshType = 'Accessory'
    if fullbase.startswith("hair_") :
        mshType = 'Hair'
    if fullbase.startswith("hat_") :
        mshType = 'Hat'
    if fullbase.startswith("latex_") :
        mshType = 'Latex'
    if fullbase.startswith("blp_") :
        mshType = 'Blueprint'
    if fullbase.startswith("bd_") :
        mshType = 'Backdrop'
    scene['mshType'] = mshType
            
    scene['hide_actor_head'] = header.hide_actor_head
    scene['hide_actor_hair'] =  header.hide_actor_hair
    scene['is_auto_animated'] = header.is_auto_animated
    scene['tex_replace_mode'] = header.tex_replace_mode
    scene['has_static_anim'] = header.has_static_anim
    scene['has_shapes'] = header.has_shapes
    scene['has_childmesh'] = header.has_childmesh
    scene['has_rooms'] = header.has_rooms
    print('Loading images')
    print('materials = ',bpy.data.materials.keys())
    image_list = []
    for imageid in range(header.image_count):
        imagefile = read_nts(stream).strip().strip('\x00')#.replace(".dds",".png")
        testimage = os.path.join(filepath, imagefile).strip().strip('\x00')
        
        if os.path.isfile(testimage):
            imageload = testimage
        else:
            imageload = get_image_path(imagedir, imagefile)
        if imageload:

            bpy.ops.image.open(filepath=imageload)
        else:
            print('image load failed.  Defaulting...')
            bpy.ops.image.new(name=imagefile,width=256,height=256)
            print("Image", imagefile)
        image_list += [imagefile]
        images = bpy.data.images

        #print "Image %s:"%imageid, imagefile
    print('images loaded: %.4f sec.' % (time()-prev_time))
    print('materials = ',bpy.data.materials.keys())
    prev_time = time()
    texkeys = bpy.data.textures.keys()
    matno = len(bpy.data.materials.keys())    
    for materialid in range(header.material_count):
        mat_id = materialid + matno
        mat = bpy.data.materials.new('Material.%03d'%mat_id)
        mat.use_face_texture = True
        mat.use_uv_project = False
        mat.use_raytrace = False
        print('mat.name = ',mat.name)
        mat.specular_intensity = 0
        tmat = material(stream)
        mat.diffuse_color = [tmat.color_R/255.0,tmat.color_G/255.0,
                             tmat.color_B/255.0]
# Diffuse:
        tex = bpy.data.textures.new('Diffuse',type='IMAGE')
        print('tmat.use_alpha = ',tmat.use_alpha)
        if tmat.use_alpha:
            tex.use_alpha = True
        if tmat.map0 != -1: 
            image_name = image_list[tmat.map0]
            print('Diffuse: image name = ',image_name)
            for img in images:
                if img.name == image_name:
                    tex.image = img
            tex.image.use_alpha = tmat.use_alpha
            mat.active_texture = tex
            if tmat.use_alpha == 1:
                mat.use_transparency = True
                mat.alpha = 0
                bpy.data.materials[mat.name].texture_slots[0].use_map_alpha = True
        else:
            tex.type = 'NONE'
            mat.use_textures[0] = False
# Reflection:
        tex = bpy.data.textures.new('Reflection',type = 'IMAGE')
        mat.texture_slots.add()
        mat.active_texture_index = 1
        mat.active_texture = tex
        if tmat.map1 != -1:
            image_name =  image_list[tmat.map1]
            for img in images:
                if img.name == image_name:
                    tex.image = img
            tex.intensity = .25
        else:
            tex.type = 'NONE'
            mat.use_textures[1] = False
# Lightmap:
        tex = bpy.data.textures.new('Lightmap',type = 'IMAGE')
        tex.filter_type = 'BOX'
        mat.texture_slots.add()
        mat.active_texture_index = 2
        mat.texture_slots[2].blend_type = 'MULTIPLY'
        mat.active_texture = tex
        mat.use_textures[2] = False
        print('mat.name = ',mat.name)
        print('blend type = ',mat.texture_slots[2].blend_type)
        if tmat.map2 != -1:
            image_name = image_list[tmat.map2]
            for img in images:
                if img.name == image_name:
                    tex.image = img
        else:
            tex.type = 'NONE'
            mat.use_textures[2] = False
# Specular:
        tex = bpy.data.textures.new('Specular',type='IMAGE')
        mat.texture_slots.add()
        mat.active_texture_index = 3
        mat.active_texture = tex
        if tmat.map3 != -1:
            image_name = image_list[tmat.map3]
            for img in images:
                if img.name == image_name:
                    tex.image = img
        else:
            tex.type = 'NONE'
            mat.use_textures[3] =False
        
        ## Populate Material ID Properties
        if ('TheMovies' in mat) == False:
            mat['TheMovies'] = 'Material'
        mat['doublesided'] = tmat.doublesided
        mat['floor_shadow_tex'] = tmat.floor_shadow_tex
        mat['alpha_separate'] =  tmat.alpha_separate
        mat['wrap_U'] = tmat.wrap_U
        mat['wrap_V'] = tmat.wrap_V
        mat['enable_alpha_test'] = tmat.enable_alpha_test
        mat['glass'] = tmat.glass
        mat['water'] = tmat.water
        mat['still_water'] = tmat.still_water
        mat['alpha_per_vertex'] = tmat.alpha_per_vertex
        mat['alphaenvmap'] = tmat.alphaenvmap        
        mat['invisible'] = tmat.invisible
        mat['not_z_write'] = tmat.not_z_write
        mat['self_lit'] = tmat.self_lit
        mat['no_floor_shadow'] = tmat.no_floor_shadow
        mat['no_delaydraw_transp'] = tmat.no_delaydraw_transp
        mat['tri_sort'] = tmat.tri_sort
        mat['additive'] = tmat.additive
        mat['scroll_U'] = tmat.scroll_U
        mat['scroll_V'] = tmat.scroll_V
        mat['rot_UV'] = tmat.rot_UV

        
        material_list.append(mat)
    
    print('materials processed: %.4f sec.' % (time()-prev_time))
    prev_time = time()

    
    if header.has_clickable:
        n = "clickable"
        add_control_mesh(n, stream, 2)
    
    if header.has_collision:
        n = "collision"
        add_control_mesh(n, stream, 2)
    
    if header.has_shadow:
        n = "shadow"
        add_control_mesh(n, stream, 2)

    if header.has_z_height:
        n = "z_height"
        add_control_mesh(n, stream, 2)
    
    if header.has_neg_space:
        n = "neg_space"
        add_control_mesh(n, stream, 2)
    
    if header.has_min_outline:
        n = "min_outline"
        add_control_mesh(n, stream, 2)
    
    if header.has_lot_boundary:
        n = "lot_boundary"
        add_control_mesh(n, stream, 2)
    
    for roomid in range(header.room_count):
        roomname = read_nts(stream)
        n = roomname
        add_control_mesh(n, stream, 3)
        
    print('controlmeshes done: %.4f sec.' % (time()-prev_time))
    prev_time = time()

    object_names = []
    group_names = []
    bones = {}
    for groupid in range(header.group_count):
        h = group_header(stream)
        group_name = "%02d"%groupid
        piv = bpy.data.objects.new(group_name, None)
        # piv.matrix_world = h.pivot
        m = h.pivot
        mat = Matrix([[m[0][0],m[0][1],m[0][2]],
                      [m[1][0],m[1][1],m[1][2]],
                      [m[2][0],m[2][1],m[2][2]]]).transposed()
        piv.rotation_euler = mat.to_euler()
        piv.location = (m[3][0],m[3][1],m[3][2])
#        piv.rotation_euler = mathutils.Euler([0,0,0])
        # piv.setMatrix(h.pivot)
        group_names.append(group_name)
        # scene.collection.objects.link(piv)
        group_obj = bpy.data.groups.get(group_name)
        if group_obj == None:
            group_obj = bpy.data.groups.new(group_name)
#        try:
#            group_obj = bpy.data.groups.get(group_name) 
#        except:
#            group_obj = bpy.data.groups.new(group_name)
        group_obj.objects.link(piv)
        scene.objects.link(piv)
#        bpy.context.scene.collection.objects.link(piv)
#        scene.objects.link(piv)
        
        ## Populate Group ID Properties
        if ('TheMovies' in piv) == False:
            piv['TheMovies'] = 'Group'
        piv['has_transanim'] = h.has_transanim
        piv['has_transanim'] = h.has_transanim
        piv['is_land'] = h.is_land
        piv['is_a_carbody'] = h.is_a_carbody
        piv['hide_reflection'] = h.hide_reflection
        piv['has_hidden'] = h.has_hidden
        piv['hidden_on'] = h.hidden_on
        piv['hidden_off'] = h.hidden_off

        scene = bpy.data.scenes["Scene"]
        for meshid in range(h.mesh_count):
            mh = mesh_header(stream)
            m, b, nd = get_mesh("Mesh.%03d"%meshid, mh, stream)           
            mn = "%s.%03d"%(group_name,meshid)
            obj = bpy.data.objects.new(mn,m)
            scene.objects.link(obj)
            object_names.append(mn)
            m2 = scene.objects.get(mn).data
            
            ## Populate Mesh ID Properties
            if('TheMovies' in m2) == False:
                m2['TheMovies'] = 'Mesh'
            m2['has_floor_reflections'] = mh.has_floor_reflections
            m2['no_outline'] = mh.no_outline
            m2['is_landscape'] = mh.is_landscape
            m2['has_neckconnect'] = mh.has_neckconnect #user has to change this to override
            m2['accepts_actor_shadow'] = mh.accepts_actor_shadow
            m2['is_minutehand'] = mh.is_minutehand
            m2['is_hourhand'] = mh.is_hourhand
            m2['static_backdrop'] = mh.static_backdrop
            m2['unk_flag'] = mh.unk_flag
            m2['has_meshid'] = mh.has_meshid
            m2['bone_per_vertex'] = mh.bone_per_vertex
            if mh.has_meshid:
                m2['indexid'] = mh.indexid
                m2['vertexid'] = mh.vertexid
                m2['skeletonid'] = mh.skeletonid
                m2['generate_new_id'] = 1
            else:
                m2['indexid'] = 0
                m2['vertexid'] = 0
                m2['skeletonid'] = 0
                m2['generate_new_id'] = 0
            
            if b != {}:
                if(group_name in bones) == False:
                    bones[group_name] = {}
                bones[group_name][mn] = b
            

            
            print('mesh %s: %.4f sec.' % (mn, time()-prev_time))
            prev_time = time()
            
            if mh != None:
                if mh.has_neckconnect:
                    if mh.neckconnect_count > 0:
                        MVGr = obj.data
                        bpy.context.scene.objects.active = obj
                        ng = bpy.context.active_object.vertex_groups.new(name='neckconnect')
                        ng.add(nd,1.0,'ADD') 
    if header.has_bones:
        h = skeleton_header(stream)
        bpy.ops.object.add(type='ARMATURE')
        arm = bpy.data.objects['Armature']
        arm.name = str(h.rig_id)
        bpy.context.scene.objects.active = arm
        arm.location = (0,0,0)
        bpy.ops.object.mode_set(mode='EDIT', toggle=False)
        if str(h.rig_id) in arm:
            arm['skeletonType'] = tmConst.RIG_IDS[h.rig_id]
        else: 
            arm['skeletonType'] = 'Unknown'
        
        a = arm.data

        #arm.link(a) #< - deprecated
#        a.makeEditable()
        edit_bones = arm.data.edit_bones
#        a.envelopes = False
#        a.autoIK = True
        a.draw_type = 'STICK'
        
        bone_list=[]
        safebones = []
        for boneid in range(h.bone_count):
            tmb = bone(stream)
            #Store bone-order data...
            if ('BoneOrder' in arm) ==False:
                arm['BoneOrder']= {}
            arm['BoneOrder'][str(boneid)]=(tmb.bone_name).rstrip('\00')
            arm[str(boneid)] =  arm['BoneOrder'][str(boneid)]
            tmb.matrix.transpose()
            pos = tmb.matrix.to_translation()
            axis, roll = mat3_to_vec_roll(tmb.matrix.to_3x3())
            b = edit_bones.new(tmb.bone_name)
            b.head = pos
            rscale = Matrix([(.125,0,0),(0,.125,0),(0,0,.125)])
            b.tail = pos + rscale*axis
            b.roll = roll
            
#pos = mymatrix.to_translation()
#axis, roll = mat3_to_vec_roll(mymatrix.to_3x3())
#bone = armature.edit_bones.new('name')
#bone.head = pos
#bone.tail = pos + axis
#bone.roll = roll            
            bone_list.append(b)
            safebones.append(b.name)
            
            if tmb.parent != -1:
                bparent = bone_list[tmb.parent]
                if bparent.tail != b.head:
#                    bangle = AngleBetweenVecs(bparent.tail - bparent.head, bparent.tail - b.head)
                    bangle = (bparent.tail-bparent.head).angle(bparent.tail - b.head)*(180.0/math.pi)
                    if bangle <0.0009 or (180-bangle)<.0009:
                        if (bparent.head-b.head).magnitude >.0009:
                            bone_list[tmb.parent].tail = b.head
#                            b.options = Blender.Armature.CONNECTED
                            b.use_connect = True
                else:
                    b.use_connect = True
                b.parent = bone_list[tmb.parent]
#            (v1,v2,v3) = (tmb.matrix[0],tmb.matrix[1],tmb.matrix[2])
#            b.tail = b.head + Vector((.125*v2[0],.125*v2[1],.125*v2[2]))     
##            a.bones[b.name] = b
##        a.update()
#    a.name = 'Bones'
        bpy.ops.object.mode_set(mode='OBJECT')
        print('bones added: %.4f sec.' % (time()-prev_time))
        prev_time = time()
        ob = bpy.data.objects['00.000']
        # Since (for some unknown reason), the bone_list was being corrupted,
        # I've changed this to save the names, as that's the only bit we
        # seem to be using here.
        bpy.ops.object.select_all(action='DESELECT')
        for gn in bones.keys():
            for mn in bones[gn].keys():
                for boneid in bones[gn][mn].keys():
                    bname = safebones[boneid]
#                    ob = Blender.Object.Get(mn)
                    ob = bpy.data.objects[mn]
#                    m = ob.getData()
                    m = ob.data
                    ob.select = True
                    bpy.context.scene.objects.active = ob
                    ng = bpy.context.active_object.vertex_groups.new(name=bname)                   
                    for w in bones[gn][mn][boneid].keys():
                        if w < 0 or w > 1:
                            print("FAIL! Weight out of range:", w)
                        v_list = bones[gn][mn][boneid][w]
                        ng.add(v_list,w,'ADD')
        arm.select = True
        bpy.context.scene.objects.active = arm
        bpy.ops.object.parent_set(type='ARMATURE')           
        print('bones weighted: %.4f sec.' % (time()-prev_time))
        prev_time = time()
    bpy.ops.object.select_all(action='DESELECT')                  
    # Groups
#           if (ob.name in bpy.data.groups.keys()) == False:
#                bpy.data.groups.new(ob.name)
    name_trans = {}
    for groupid in range(header.group_count):
        group_name = "%02d"%groupid
        gread = group_name
        if header.has_static_anim or header.has_named_groups: 
            gread = read_nts(stream).strip('\x00').strip('\x00')
        name_trans[group_name] = gread
        gr = bpy.data.objects.get(group_names[groupid])
        gr.name = group_name + "." + gread
        gr['grpName'] = gread
        group_obj = bpy.data.groups.get(group_names[groupid])
        group_obj.name = group_name + "." + gread
        name_trans[group_name] = gr.name
    loop = 0
    for mn in object_names:
        loop = loop + 1
        gn, tail = mn.split(".",1)
        if gn in name_trans:
            gr = bpy.data.objects[name_trans[gn]]
            group_obj = bpy.data.groups.get(name_trans[gn])
            mesh  = bpy.data.objects[mn]
            new_name = "Object.%s"%(tail)
            mesh.name = new_name
#            print('Group object is None.')
            # if mesh is child of Armature, make Armature the child of Group, otherwise
            # make mesh child of Group
            if mesh.parent == None:
# bpy.ops.object.parent_set(type='OBJECT',xmirror=False,keep_transform=True)
                mesh.select = True
                gr.select = True
                grp_obj = bpy.data.objects[group_obj.name]
                bpy.context.scene.objects.active = grp_obj
                bpy.ops.object.parent_set(type='OBJECT',xmirror=False,keep_transform=True)
                mesh.select = False
                gr.select = False
            elif mesh.parent.type == 'ARMATURE':
                mesh.parent.select = True
                mesh.select = False
                bpy.context.scene.objects.active = gr
                bpy.ops.object.parent_set(type='OBJECT',xmirror=False,keep_transform=True)
            group_obj.objects.link(mesh)
    print('Groups finalized: %.4f sec.' % (time()-prev_time))
    prev_time = time()
# Anchors go here
#group_anchors = bpy.data.groups.get('anchors')
#if group_anchors == None:
#    group_anchors = bpy.data.groups.new('anchors')
    if header.anchor_count > 0:
        group_anchors = bpy.data.groups.get('anchors')
        if group_anchors == None:
            group_anchors = bpy.data.groups.new('anchors')
        anchors = []
        anchParent = bpy.data.objects.get('Anchors')
        if anchParent == None:
            anchParent = bpy.data.objects.new('Anchors',None)
            group_anchors.objects.link(anchParent)
#    group_anchors.objects.link(anchParent)
        scene.objects.link(anchParent)
    for anchorid in range(header.anchor_count):
        ca = anchor(stream)
        anch = bpy.data.objects.new(ca.anchor_name,None)
        scene.objects.link(anch)
        m = Matrix([[ca.m11,ca.m12,ca.m13,0],
                    [ca.m21,ca.m22,ca.m23,0],
                    [ca.m31,ca.m32,ca.m33,0],
                    [ca.offset_x,ca.offset_y,ca.offset_z,1]])
        anch.matrix_local = m.transposed()
        anch.location = [ca.offset_x,ca.offset_y,ca.offset_z]
        anch.parent = anchParent
        group_anchors.objects.link(anch)
        anchors.append(anch)
        anch['ancName'] = ca.anchor_name
    #Convex Hull goes here -- but...we're skipping it...    
    if header.has_convex_hull:
        chullsize = read_item("L", stream)
        print('convex hull: %s bytes, skipping.' %chullsize)
        stream.seek(chullsize-4,1)
        

            
    print('shapes processed: %.4f sec.' % (time()-prev_time))
#    prev_time = time()
                
    stream.close()
    
#    if in_editmode:
#        Window.EditMode(1)
#    scene.setLayers([1])
#    Blender.Redraw()
# Let's try patching up the groups here.

#    for ob in bpy.data.objects:
#        if ob.type == 'EMPTY':
#            if (ob.name in bpy.data.groups.keys()) == False:
#                bpy.data.groups.new(ob.name)
#                groups = bpy.data.groups
#                print('groups = ',groups)
#                group = groups[ob.name]
#                group.objects.link(ob)
#        if ob.type == 'MESH':
#            group.objects.link(ob)

#group.objects.link(obj)            
    
    print('finished importing: "%s" in %.4f sec.' % (filename, (time()-start_time)))









def read_mesh(context, filepath):
    load_msh(filepath)
        
    return {'FINISHED'}


class TheMoviesMeshImporter(Operator, ImportHelper):
    """Import a TM msh object"""
    bl_idname = "import_test.tm_msh"  # important since its how bpy.ops.import_test.some_data is constructed
    bl_label = "Import TM MSH"
    bl_options = {'REGISTER','UNDO'}

    # ImportHelper mixin class uses this
    filename_ext = ".msh"

    filter_glob = StringProperty(
        default="*.msh",
        options={'HIDDEN'},
        maxlen=255,  # Max internal buffer length, longer would be clamped.
    )

    def execute(self, context):
        return read_mesh(context, self.filepath)

# Only needed if you want to add into a dynamic menu
def menu_func_import(self, context):
    self.layout.operator(TheMoviesMeshImporter.bl_idname, text="Import The Movies Mesh (*.msh)")

def register():
    bpy.utils.register_class(TheMoviesMeshImporter)
#    bpy.types.TOPBAR_MT_file_import.append(menu_func_import)
    bpy.types.INFO_MT_file_import.append(menu_func_import)    
def unregister():
    bpy.utils.unregister_class(TheMoviesMeshImporter)
#    bpy.types.TOPBAR_MT_file_import.remove(menu_func_import)
    bpy.types.INFO_MT_file_import.append(menu_func_import)
#if __name__ == "__main__":
#    register()
def mshin():
#    register()
    # test call
    bpy.ops.import_test.tm_msh('INVOKE_DEFAULT')
#register()
if __name__ == "__main__":
    register()
#    bpy.ops.import_test.tm_msh('INVOKE_DEFAULT')
#    filepath = 'c:\TMBlender\data\meshes\p_heli_tran.msh'
#    load_msh(filepath)
#
# bpy.ops.object.armature_add()   creates an Armature named Armature
# obs = bpy.data.objects.keys()   Gives list of names 
# obs = bpy.data.objects          List of objects
# ob = obs[N] where N is the index of the armature in the list of objects
# ob.name = 'new name for armature'
#
